# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance, too-many-arguments

"""Data structures for music service items

The basis for this implementation is this page in the Sonos API
documentation: http://musicpartners.sonos.com/node/83

A note about naming. The Sonos API uses camel case with starting lower
case. These names have been adapted to match general Python class
naming conventions.

MediaMetadata:
    Track
    Stream
    Show
    Other

MediaCollection:
    Artist
    Album
    Genre
    Playlist
    Search
    Program
    Favorites
    Favorite
    Collection
    Container
    AlbumList
    TrackList
    StreamList
    ArtistTrackList
    Other

NOTE: "Other" is allowed under both.

Class overview:

+----------------+   +----------------+   +---------------+
|MetadataDictBase+-->+MusicServiceItem+-->+MediaCollection|
+-----+----------+   +--------+-------+   +---------------+
      |                       |
      |                       |     +------------------+
      |                       +---->+  MediaMetadata   |
      |                             |                  |
      |                             | +-------------+  |
      +------------------------------>+TrackMetadata|  |
      |                             | +-------------+  |
      |                             |                  |
      |                             | +--------------+ |
      +------------------------------>+StreamMetadata| |
                                    | +--------------+ |
                                    |                  |
                                    +------------------+


"""
from urllib.parse import quote as quote_url

import logging
from collections import OrderedDict
from ..data_structures import DidlResource, DidlItem, SearchResult
from ..utils import camel_to_underscore


_LOG = logging.getLogger(__name__)
_LOG.addHandler(logging.NullHandler())


# For now we generate classes dynamically. This is shorter, but
# provides no custom documentation for all the different types.
CLASSES = {}


def get_class(class_key):
    """Form a music service data structure class from the class key

    Args:
        class_key (str): A concatenation of the base class (e.g. MediaMetadata)
            and the class name

    Returns:
        class: Subclass of MusicServiceItem
    """
    if class_key not in CLASSES:
        for basecls in (MediaMetadata, MediaCollection):
            if class_key.startswith(basecls.__name__):
                # So MediaMetadataTrack turns into MSTrack
                class_name = "MS" + class_key.replace(basecls.__name__, "")
                CLASSES[class_key] = type(class_name, (basecls,), {})
                _LOG.debug("Class %s created", CLASSES[class_key])
    return CLASSES[class_key]


def parse_response(service, response, search_type):
    """Parse the response to a music service query and return a SearchResult

    Args:
        service (MusicService): The music service that produced the response
        response (OrderedDict): The response from the soap client call
        search_type (str): A string that indicates the search type that the
            response is from

    Returns:
        SearchResult: A SearchResult object
    """
    _LOG.debug(
        'Parse response "%s" from service "%s" of type "%s"',
        response,
        service,
        search_type,
    )
    items = []
    # The result to be parsed is in either searchResult or getMetadataResult
    if "searchResult" in response:
        response = response["searchResult"]
    elif "getMetadataResult" in response:
        response = response["getMetadataResult"]
    else:
        raise ValueError(
            '"response" should contain either the key '
            '"searchResult" or "getMetadataResult"'
        )

    # Form the search metadata
    search_metadata = {
        "number_returned": response["count"],
        "total_matches": None,
        "search_type": search_type,
        "update_id": None,
    }

    for result_type in ("mediaCollection", "mediaMetadata"):
        # Upper case the first letter (used for the class_key)
        result_type_proper = result_type[0].upper() + result_type[1:]
        raw_items = response.get(result_type, [])
        # If there is only 1 result, it is not put in an array
        if isinstance(raw_items, OrderedDict):
            raw_items = [raw_items]

        for raw_item in raw_items:
            # Form the class_key, which is a unique string for this type,
            # formed by concatenating the result type with the item type. Turns
            # into e.g: MediaMetadataTrack
            class_key = result_type_proper + raw_item["itemType"].title()
            cls = get_class(class_key)
            items.append(cls.from_music_service(service, raw_item))
    return SearchResult(items, **search_metadata)


def form_uri(item_id, service, is_track):
    """Form and return a music service item uri

    Args:
        item_id (str): The item id
        service (MusicService): The music service that the item originates from
        is_track (bool): Whether the item_id is from a track or not

    Returns:
        str: The music service item uri
    """
    if is_track:
        uri = service.sonos_uri_from_id(item_id)
    else:
        uri = "x-rincon-cpcontainer:" + item_id
    return uri


# Type Helper
BOOL_STRS = {"true", "false"}


def bool_str(string):
    """Returns a boolean from a string imput of 'true' or 'false'"""
    if string not in BOOL_STRS:
        raise ValueError('Invalid boolean string: "{}"'.format(string))
    return string == "true"


# Music Service item base classes
class MetadataDictBase:
    """Class used to parse metadata from kwargs"""

    # The following two fields should be overwritten in subclasses

    # _valid_fields is a set of valid fields
    _valid_fields = {}

    # _types is a dict of fields with non-string types and their convertion
    # callables
    _types = {}

    def __init__(self, metadata_dict):
        """Initialize local variables"""
        _LOG.debug("MetadataDictBase.__init__ with: %s", metadata_dict)
        for key in metadata_dict:
            # Check for invalid fields
            if key not in self._valid_fields:
                message = '%s instantiated with invalid field "%s" and value: "%s"'
                # Really wanted to raise exceptions here, but as it
                # turns out I have already encountered invalid fields
                # from music services.
                _LOG.debug(message, self.__class__, key, metadata_dict[key])

        # Convert names and create metadata dict
        self.metadata = {}
        for key, value in metadata_dict.items():
            if key in self._types:
                convertion_callable = self._types[key]
                value = convertion_callable(value)
            self.metadata[camel_to_underscore(key)] = value

    def __getattr__(self, key):
        """Return item from metadata in case of unknown attribute"""
        try:
            return self.metadata[key]
        except KeyError as error:
            message = 'Class {} has no attribute "{}"'
            raise AttributeError(
                message.format(self.__class__.__name__, key)
            ) from error


class MusicServiceItem(MetadataDictBase):
    """A base class for all music service items"""

    # See comment in MetadataDictBase for explanation of these two attributes
    _valid_fields = {}
    _types = {}

    def __init__(
        self,
        item_id,
        desc,  # pylint: disable=too-many-arguments
        resources,
        uri,
        metadata_dict,
        music_service=None,
    ):
        """Init music service item

        Args:
            item_id (str): This is the Didl compatible id NOT the music item id
            desc (str): A DIDL descriptor, default ``'RINCON_AssociatedZPUDN'
            resources (list): List of DidlResource
            uri (str): The uri for the location of the item
            metdata_dict (dict): Mapping of metadata
            music_service (MusicService): The MusicService instance the item
                originates from
        """
        _LOG.debug(
            "%s.__init__ with item_id=%s, desc=%s, resources=%s, "
            "uri=%s, metadata_dict=..., music_service=%s",
            self.__class__.__name__,
            item_id,
            desc,
            resources,
            uri,
            music_service,
        )
        super().__init__(metadata_dict)
        self.item_id = item_id
        self.desc = desc
        self.resources = resources
        self.uri = uri
        self.music_service = music_service

    @classmethod
    def from_music_service(cls, music_service, content_dict):
        """Return an element instantiated from the information that a music
        service has (alternative constructor)

        Args:
            music_service (MusicService): The music service that content_dict
                originated from
            content_dict (OrderedDict): The data to instantiate the music
                service item from

        Returns:
            MusicServiceItem: A MusicServiceItem instance
        """
        # Form the item_id
        quoted_id = quote_url(content_dict["id"].encode("utf-8"))
        # The hex prefix remains a mistery for now
        item_id = "0fffffff{}".format(quoted_id)
        # Form the uri
        is_track = cls == get_class("MediaMetadataTrack")
        uri = form_uri(item_id, music_service, is_track)
        # Form resources and get desc
        resources = [DidlResource(uri=uri, protocol_info="DUMMY")]
        desc = music_service.desc
        return cls(
            item_id, desc, resources, uri, content_dict, music_service=music_service
        )

    def __str__(self):
        """Return custom string representation"""
        title = self.metadata.get("title")
        str_ = '<{} title="{}">'
        return str_.format(self.__class__.__name__, title)

    def to_element(self, include_namespaces=False):
        """Return an ElementTree Element representing this instance.

        Args:
            include_namespaces (bool, optional): If True, include xml
                namespace attributes on the root element

        Return:
            ~xml.etree.ElementTree.Element: The (XML) Element representation of
                this object
        """
        # We piggy back on the implementation in DidlItem
        didl_item = DidlItem(
            title="DUMMY",
            # This is ignored. Sonos gets the title from the item_id
            parent_id="DUMMY",  # Ditto
            item_id=self.item_id,
            desc=self.desc,
            resources=self.resources,
        )
        return didl_item.to_element(include_namespaces=include_namespaces)


class TrackMetadata(MetadataDictBase):
    """Track metadata class"""

    # _valid_fields is a set of valid fields
    _valid_fields = {
        "artistId",
        "artist",
        "composerId",
        "composer",
        "albumId",
        "album",
        "albumArtURI",
        "albumArtistId",
        "albumArtist",
        "genreId",
        "genre",
        "duration",
        "canPlay",
        "canSkip",
        "canAddToFavorites",
        "rating",
        "trackNumber",
        "isFavorite",
    }
    # _types is a dict of fields with non-string types and their
    # convertion callables
    _types = {
        "duration": int,
        "canPlay": bool_str,
        "canSkip": bool_str,
        "canAddToFavorites": bool_str,
        "rating": int,
        "trackNumber": int,
        "isFavorite": bool_str,
    }


class StreamMetadata(MetadataDictBase):
    """Stream metadata class"""

    # _valid_fields is a set of valid fields
    _valid_fields = {
        "currentHost",
        "currentShowId",
        "currentShow",
        "secondsRemaining",
        "secondsToNextShow",
        "bitrate",
        "logo",
        "hasOutOfBandMetadata",
        "description",
        "isEphemeral",
    }
    # _types is a dict of fields with non-string types and their
    # convertion callables
    _types = {
        "secondsRemaining": int,
        "secondsToNextShow": int,
        "bitrate": int,
        "hasOutOfBandMetadata": bool_str,
        "isEphemeral": bool_str,
    }


class MediaMetadata(MusicServiceItem):
    """Base class for all media metadata items"""

    # _valid_fields is a set of valid fields
    _valid_fields = {
        "id",
        "title",
        "mimeType",
        "itemType",
        "displayType",
        "summary",
        "trackMetadata",
        "streamMetadata",
        "dynamic",
    }
    # _types is a dict of fields with non-string types and their
    # convertion callables
    _types = {
        "trackMetadata": TrackMetadata,
        "streamMetadata": StreamMetadata,
        # We ignore types on the dynamic field
        # 'dynamic': ???,
    }


class MediaCollection(MusicServiceItem):
    """Base class for all mediaCollection items"""

    # _valid_fields is a set of valid fields
    _valid_fields = {
        "id",
        "title",
        "itemType",
        "displayType",
        "summary",
        "artistId",
        "artist",
        "albumArtURI",
        "canPlay",
        "canEnumerate",
        "canAddToFavorites",
        "containsFavorite",
        "canScroll",
        "canSkip",
        "isFavorite",
    }

    # _types is a dict of fields with non-string types and their
    # convertion callables
    _types = {
        "canPlay": bool_str,
        "canEnumerate": bool_str,
        "canAddToFavorites": bool_str,
        "containsFavorite": bool_str,
        "canScroll": bool_str,
        "canSkip": bool_str,
        "isFavorite": bool_str,
    }



"""py.test hooks.

Add the --ip command line option, and skip all tests marked the with
'integration' marker unless the option is included
"""
from os import path
import json
import codecs


import pytest

THISDIR = path.dirname(path.abspath(__file__))


def pytest_addoption(parser):
    """Add the --ip commandline option"""
    parser.addoption(
        "--ip",
        type=str,
        default=None,
        action="store",
        dest="IP",
        help="the IP address for the zone to be used for the integration tests",
    )


def pytest_runtest_setup(item):
    """Skip tests marked 'integration' unless an ip address is given."""
    if "integration" in item.keywords and not item.config.getoption("--ip"):
        pytest.skip("use --ip and an ip address to run integration tests.")


class Helpers:
    """Test helper functions"""

    @staticmethod
    def compare_xml(element1, element2):
        # print()
        # print(element1.tag, "TAG", element2.tag)
        # print(element1.attrib, element2.attrib)
        if element1.tag != element2.tag:
            return False

        # print(element1.attrib, element2.attrib)
        if element1.attrib != element2.attrib:
            return False

        # print(len(element1), len(element2))
        if len(element1) != len(element2):
            return False

        for subelement1, subelement2 in zip(element1, element2):
            if not Helpers.compare_xml(subelement1, subelement2):
                return False

        return True


@pytest.fixture
def helpers():
    return Helpers


class DataLoader:
    """A class that loads test data"""

    def __init__(self, data_sub_dir):
        self.data_dir = path.join(THISDIR, "data", data_sub_dir)

    def load_xml(self, filename):
        """Return XML string loaded from filename under ``self.data_sub_dir``"""
        xml_string = ""
        with codecs.open(path.join(self.data_dir, filename), encoding="utf-8") as file_:
            for line in file_:
                # Allow for indenting the XML source
                xml_string += line.lstrip(" ")
        return xml_string

    def load_json(self, filename):
        """Return parsed data from json file"""
        with open(path.join(self.data_dir, filename)) as file_:
            data = json.load(file_)
        return data

    def load_xml_and_json(self, filename_base):
        """Return XML string and parsed data from ``filename_base``.xml and .json"""
        return (
            self.load_xml(filename_base + ".xml"),
            self.load_json(filename_base + ".json"),
        )



"""Test for music_services/data_structures"""


from collections import OrderedDict
import pytest
from unittest.mock import PropertyMock, Mock, patch
from pysonos.music_services import data_structures
from pysonos.data_structures import DidlResource

# DATA
RESPONSES = []
RESPONSES.append(
    OrderedDict(
        [
            (
                "searchResult",
                OrderedDict(
                    [
                        ("index", "0"),
                        ("count", "2"),
                        ("total", "17230"),
                        (
                            "mediaCollection",
                            [
                                OrderedDict(
                                    [
                                        ("id", "album/43820695"),
                                        ("itemType", "album"),
                                        ("title", "Black Mosque"),
                                        ("artist", "Black Mosque"),
                                        ("artistId", "artist/6689314"),
                                        ("canPlay", "true"),
                                        ("canEnumerate", "true"),
                                        ("canAddToFavorites", "true"),
                                        ("canSkip", "true"),
                                        (
                                            "albumArtURI",
                                            "http://resources.wimpmusic.com/images/"
                                            "2238a5cd/ed4d/4ad0/848d/40356f11bda0/"
                                            "640x640.jpg",
                                        ),
                                        ("canAddToFavorite", "true"),
                                    ]
                                ),
                                OrderedDict(
                                    [
                                        ("id", "album/50340580"),
                                        ("itemType", "album"),
                                        ("title", "Black Hippy 2"),
                                        ("artist", "Black Hippy"),
                                        ("artistId", "artist/3882538"),
                                        ("canPlay", "true"),
                                        ("canEnumerate", "true"),
                                        ("canAddToFavorites", "true"),
                                        ("canSkip", "true"),
                                        (
                                            "albumArtURI",
                                            "http://resources.wimpmusic.com/images/"
                                            "eefb1532/dfe9/"
                                            "46bd/8775/c583844bc098/640x640.jpg",
                                        ),
                                        ("canAddToFavorite", "true"),
                                    ]
                                ),
                            ],
                        ),
                    ]
                ),
            )
        ]
    )
)
RESPONSES.append(
    OrderedDict(
        [
            (
                "getMetadataResult",
                OrderedDict(
                    [
                        (
                            "mediaMetadata",
                            OrderedDict(
                                [
                                    ("id", "Track@catalog:/tracks/104655624"),
                                    ("title", "Take Me Into Your Skin"),
                                    ("itemType", "track"),
                                    ("mimeType", "audio/aac"),
                                    (
                                        "trackMetadata",
                                        OrderedDict(
                                            [
                                                (
                                                    "albumArtURI",
                                                    "http://artwork.cdn.247e.com/"
                                                    "covers/104655587/256x256",
                                                ),
                                                (
                                                    "artistId",
                                                    "Artist@catalog:/artists/219591",
                                                ),
                                                ("artist", "Trentemøller"),
                                                ("album", "The Last Resort"),
                                                ("duration", "464"),
                                                ("canPlay", "true"),
                                                ("canSkip", "true"),
                                                ("canAddToFavorites", "true"),
                                                ("trackNumber", "1"),
                                            ]
                                        ),
                                    ),
                                ]
                            ),
                        ),
                        ("count", "1"),
                        ("index", "0"),
                        ("total", "13"),
                    ]
                ),
            )
        ]
    )
)
PARSE_RESULTS = (
    {
        "number_of_results": 2,
        "type": "searchResult",
        "class_key": "MediaCollectionAlbum",
    },
    {
        "number_of_results": 1,
        "type": "getMetadataResult",
        "class_key": "MediaMetadataTrack",
    },
)


def test_get_class():
    """Test the get_class function"""
    # Test core functionality for base class MediaMetadata
    cls = data_structures.get_class("MediaMetadataTrack")
    assert cls.__name__ == "MSTrack"
    assert issubclass(cls, data_structures.MediaMetadata)

    # Test core functionality for base class Mediacolection
    cls = data_structures.get_class("MediaCollectionArtist")
    assert cls.__name__ == "MSArtist"
    assert issubclass(cls, data_structures.MediaCollection)

    # Test the caching function
    cls2 = data_structures.get_class("MediaCollectionArtist")
    assert cls is cls2

    # Asking for bad class should raise KeyError
    with pytest.raises(KeyError):
        cls = data_structures.get_class("Nonsense")


@pytest.mark.parametrize("response, correct", zip(RESPONSES, PARSE_RESULTS))
def test_parse_response(response, correct):
    """Test the parse_response function"""
    music_service = Mock()
    music_service.desc = "DESC"
    results = data_structures.parse_response(music_service, response, "albums")

    # Check the search result metadata
    response_data = response[correct["type"]]
    assert results.number_returned == response_data["count"]
    assert results.search_type == "albums"

    # Check the result
    assert len(results) == correct["number_of_results"]
    klass = data_structures.get_class(correct["class_key"])
    for result in results:
        assert isinstance(result, klass)
        assert result.music_service is music_service


def test_parse_response_bad_type():
    """Test parse reponse bad code"""
    with pytest.raises(ValueError) as exp:
        data_structures.parse_response(None, {}, "albums")
    print(exp)


def test_form_uri():
    """Test the form uri function"""
    music_service = Mock()
    music_service.sonos_uri_from_id.return_value = "99"

    # Test non track uri
    non_track_uri = data_structures.form_uri("dummy_id", None, False)
    assert non_track_uri == "x-rincon-cpcontainer:dummy_id"

    # Test track uri
    track_uri = data_structures.form_uri("dummy_id", music_service, True)
    music_service.sonos_uri_from_id.assert_called_once_with("dummy_id")


def test_bool_str():
    """Test the bool_str function"""
    assert data_structures.bool_str("true") is True
    assert data_structures.bool_str("false") is False
    with pytest.raises(ValueError):
        data_structures.bool_str("dummy")


class TestMetadataDictBase:
    """Tests for the MetadataDictBase class"""

    def test_init(self):
        """Test normal __init__ functionality; metadata stored and camel case
        names coverted

        """
        metadata_dict = {
            "superTitle": "Dummy Title",
        }
        metadata = data_structures.MetadataDictBase(metadata_dict)
        # Assert only metadata element, name conversion and value
        assert len(metadata.metadata) == 1
        assert metadata.super_title == "Dummy Title"

    def test_conversion(self):
        """Test the type conversion of fields in metadata"""
        conversion_mock = Mock()
        conversion_mock.return_value = 47

        # MetadataDictBase is meant to ve overwritten, to supply
        # valid_fields and conversion functions
        class MyClass(data_structures.MetadataDictBase):
            _types = {"trackDuration": conversion_mock}

        metadata_dict = {
            "title": "Dummy Title",
            "trackDuration": "47",
        }
        metadata = MyClass(metadata_dict)
        # Check that the duration has been properly type converted
        conversion_mock.assert_called_once_with("47")
        assert metadata.track_duration == 47
        # And that title has been left unchanged
        assert metadata.title == "Dummy Title"

    def test_get_attr(self):
        """Test the __getattr__ method"""
        metadata_dict = {
            "superTitle": "Dummy Title",
        }
        metadata = data_structures.MetadataDictBase(metadata_dict)
        # Test normal lookup
        assert metadata.super_title == "Dummy Title"
        # Test raise attribute error when the key is not in metadata
        with pytest.raises(AttributeError):
            metadata.nonexistent_key


class TestMusicServiceItem:
    """Test the MusicServiceItem class"""

    @patch("pysonos.music_services.data_structures.MetadataDictBase.__init__")
    def test_init(self, metadata_dict_base_init):
        """Test the __init__ method"""
        kwargs = {
            "item_id": "some_item_id",
            "desc": "the_desc",
            "resources": "the ressources",
            "uri": "https://the_uri",
            "metadata_dict": {"some": "dict"},
            "music_service": "the music service",
        }

        music_service_item = data_structures.MusicServiceItem(**kwargs)
        # Test call to super class init
        metadata_dict_base_init.assert_called_once_with({"some": "dict"})
        # Test that all but the metadata_dict arg have been set as
        # attributes with the same names as the arguments
        kwargs.pop("metadata_dict")
        for arg_name, arg_value in kwargs.items():
            assert getattr(music_service_item, arg_name) == arg_value

    @patch("pysonos.music_services.data_structures.MusicServiceItem.__init__")
    @patch("pysonos.music_services.data_structures.form_uri")
    @patch("pysonos.music_services.data_structures.DidlResource")
    def test_from_music_service(self, didl_resource, form_uri, music_service_init):
        """Test th from music service class method"""
        # Setup mock music service with mocked desc property
        ms = Mock()
        desc = PropertyMock(return_value="fake_desc")
        type(ms).desc = desc

        # Setup content dict
        id_ = "fakeid1234"
        item_id = "0fffffff" + id_
        content_dict = {"id": id_, "title": "fake title"}

        # Setup return values of mocks
        didl_resource.return_value = Mock()
        form_uri.return_value = "x-rincon-whatever:" + id_
        music_service_init.return_value = None

        # Call the class method and assert init called
        data_structures.MusicServiceItem.from_music_service(ms, content_dict)
        music_service_init.assert_called_once_with(
            item_id,
            "fake_desc",
            [didl_resource.return_value],
            form_uri.return_value,
            content_dict,
            music_service=ms,
        )
        form_uri.assert_called_once_with(item_id, ms, False)

    def test_str_(self):
        """Test the __str__ method"""
        content_dict = {"title": "fake title"}
        item = data_structures.MusicServiceItem(
            "fake_id", "desc", "ressources", "uri", content_dict
        )
        assert item.__str__() == '<MusicServiceItem title="fake title">'

    @patch("pysonos.music_services.data_structures.DidlItem.to_element")
    def test_to_element(self, didl_item_to_element):
        """Test the to_element method"""
        didl_item_to_element.return_value = object()
        content_dict = {"title": "fake title"}
        item = data_structures.MusicServiceItem(
            "fake_id", "desc", "ressources", "uri", content_dict
        )
        assert item.to_element() == didl_item_to_element.return_value



"""A SoCo fork with fixes for Home Assistant."""

# There is no need for all strings here to be unicode, and Py2 cannot import
# modules with unicode names so do not use from __future__ import
# unicode_literals
# https://github.com/SoCo/SoCo/issues/98
#


import logging

from .core import SoCo
from .discovery import discover, discover_thread
from .exceptions import SoCoException, UnknownSoCoException

# Will be parsed by setup.py to determine package metadata
__author__ = "Anders Melchiorsen <amelchio@nogoto.net>"
# Please increment the version number and add the suffix "-dev" after
# a release, to make it possible to identify in-development code
__version__ = "0.0.52"
__website__ = "https://github.com/amelchio/pysonos"
__license__ = "MIT License"

# You really should not `import *` - it is poor practice
# but if you do, here is what you get:
__all__ = [
    "discover",
    "discover_thread",
    "SoCo",
    "SoCoException",
    "UnknownSoCoException",
]

# http://docs.python.org/2/howto/logging.html#library-config
# Avoids spurious error messages if no logger is configured by the user

logging.getLogger(__name__).addHandler(logging.NullHandler())



# pylint: disable=star-args,no-member

"""Unit tests for the music service data structures."""


from xml.sax.saxutils import escape

import pytest

from pysonos.exceptions import DIDLMetadataError
from pysonos.ms_data_structures import (
    MSAlbum,
    MSAlbumList,
    MSArtist,
    MSArtistTracklist,
    MSCollection,
    MSFavorites,
    MSPlaylist,
    MSTrack,
)
from pysonos.xml import XML

##############################################################################
# Example XML and the content dict to compare with for MS items              #
##############################################################################
DIDL_TEMPLATE = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
 xmlns:dc="http://purl.org/dc/elements/1.1/"
 xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
 xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">
<item id="{extended_id}" parentID="{parent_id}" restricted="true">
<dc:title>{title}</dc:title>
<upnp:class>{item_class}</upnp:class>
<desc id="cdudn"
 nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">{description}</desc>
</item>
</DIDL-Lite>
"""
DIDL_TEMPLATE = DIDL_TEMPLATE.replace("\n", "")

MS_TRACK_SEARCH_XML = """
<ns0:mediaMetadata xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>trackid_24125935</ns0:id>
<ns0:itemType>track</ns0:itemType>
<ns0:mimeType>audio/aac</ns0:mimeType>
<ns0:title>Pilgrim</ns0:title>
<ns0:trackMetadata>
<ns0:artistId>artistid_4816276</ns0:artistId>
<ns0:artist>MØ</ns0:artist>
<ns0:composerId>artistid_4816276</ns0:composerId>
<ns0:composer>MØ</ns0:composer>
<ns0:albumId>albumid_24125922</ns0:albumId>
<ns0:album>Nytårsfesten 2014</ns0:album>
<ns0:albumArtistId>artistid_4816276</ns0:albumArtistId>
<ns0:albumArtist>MØ</ns0:albumArtist>
<ns0:duration>231</ns0:duration>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/sca/imscale?h=90&amp;
w=90&amp;img=/content/sg19/vd00/music/prod/sonyddex/1386109801/
A10301A00030834072_20131203071914171/resources/A10301A00030834072_T-10764
_Image.jpg</ns0:albumArtURI>
<ns0:canPlay>true</ns0:canPlay>
<ns0:canSkip>true</ns0:canSkip>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
</ns0:trackMetadata>
</ns0:mediaMetadata>
"""
MS_TRACK_SEARCH_XML = MS_TRACK_SEARCH_XML.replace("\n", "")
MS_TRACK_SEARCH_DICT = {
    "can_add_to_favorites": True,
    "composer": "M\xd8",
    "album_id": "albumid_24125922",
    "can_skip": True,
    "uri": "x-sonos-http:trackid_24125935.mp4?sid=20&flags=32",
    "item_id": "trackid_24125935",
    "item_type": "track",
    "extended_id": "00030020trackid_24125935",
    "duration": 231,
    "can_play": True,
    "composer_id": "artistid_4816276",
    "description": "SA_RINCON5127_4542255535",
    "album": "Nyt\xe5rsfesten 2014",
    "title": "Pilgrim",
    "artist": "M\xd8",
    "album_artist_id": "artistid_4816276",
    "album_art_uri": "http://varnish01.music.aspiro.com/sca/imscale?"
    "h=90&w=90&img=/content/sg19/vd00/music/prod/sonyddex"
    "/1386109801/A10301A00030834072_20131203071914171/"
    "resources/A10301A00030834072_T-10764_Image.jpg",
    "album_artist": "M\xd8",
    "parent_id": "00020064tracksearch:pilgrim",
    "service_id": 20,
    "artist_id": "artistid_4816276",
    "mime_type": "audio/aac",
}
MS_ALBUM_SEARCH_XML = """
<ns0:mediaCollection readOnly="true"
 xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>albumid_5738780</ns0:id>
<ns0:itemType>album</ns0:itemType>
<ns0:title>Greatest De Unge År</ns0:title>
<ns0:artist>tv·2</ns0:artist>
<ns0:canPlay>true</ns0:canPlay>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/sca/imscale?h=90&amp;w=90
&amp;img=/content/music8/prod/sonybmg/content/00000000000002304180/000/000/
000/000/056/501/97/00000000000005650197-1000x1000_72dpi_RGB_100Q.jpg
</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_ALBUM_SEARCH_XML = MS_ALBUM_SEARCH_XML.replace("\n", "")
MS_ALBUM_SEARCH_DICT = {
    "can_add_to_favorites": True,
    "description": "SA_RINCON5127_4542255535",
    "artist": "tv\xb72",
    "title": "Greatest De Unge \xc5r",
    "album_art_uri": "http://varnish01.music.aspiro.com/sca/imscale?h=90&w=90"
    "&img=/content/music8/prod/sonybmg/content/"
    "00000000000002304180/000/000/000/000/056/501/97/"
    "00000000000005650197-1000x1000_72dpi_RGB_100Q.jpg",
    "uri": "x-rincon-cpcontainer:0004002calbumid_5738780",
    "parent_id": "00020064albumsearch:de unge",
    "item_type": "album",
    "extended_id": "0004002calbumid_5738780",
    "item_id": "albumid_5738780",
    "service_id": 20,
    "can_play": True,
}
MS_ARTIST_SEARCH_XML = """
<ns0:mediaCollection readOnly="true"
 xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>artistid_4761386</ns0:id>
<ns0:itemType>artist</ns0:itemType>
<ns0:title>Fritjof Såheim</ns0:title>
<ns0:artist>Fritjof Såheim</ns0:artist>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/im/im?h=42&amp;w=64&amp;
artistid=4761386</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_ARTIST_SEARCH_XML = MS_ARTIST_SEARCH_XML.replace("\n", "")
MS_ARTIST_SEARCH_DICT = {
    "can_add_to_favorites": True,
    "description": "SA_RINCON5127_4542255535",
    "artist": "Fritjof S\xe5heim",
    "title": "Fritjof S\xe5heim",
    "album_art_uri": "http://varnish01.music.aspiro.com/im/im?h=42&w=64"
    "&artistid=4761386",
    "parent_id": "00020064artistsearch:Fritjof",
    "item_type": "artist",
    "extended_id": "10050024artistid_4761386",
    "item_id": "artistid_4761386",
    "service_id": 20,
}
MS_PLAYLIST_SEARCH_XML = """
<ns0:mediaCollection readOnly="true" xmlns:ns0="http://www.sonos.com/Services/1.1">
  <ns0:id>playlistid_133fe1ff-1f16-4300-b440-fb80573f19ce</ns0:id>
  <ns0:itemType>albumList</ns0:itemType>
  <ns0:title>Kunstnerliste: Dans &amp; Lær</ns0:title>
  <ns0:artist>39 sange</ns0:artist>
  <ns0:artistId>artistid_0</ns0:artistId>
  <ns0:canPlay>true</ns0:canPlay>
  <ns0:canEnumerate>true</ns0:canEnumerate>
  <ns0:canAddToFavorites>true</ns0:canAddToFavorites>
  <ns0:albumArtURI>http://varnish01.music.aspiro.com/im/im?h=160&amp;w=240&amp;rows=1&amp;cols=2&amp;artimg&amp;uuid=133fe1ff-1f16-4300-b440-fb80573f19ce</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_PLAYLIST_SEARCH_XML = MS_PLAYLIST_SEARCH_XML.replace("\n", "")
MS_PLAYLIST_SEARCH_DICT = {
    "can_add_to_favorites": True,
    "description": "SA_RINCON5127_4542255535",
    "artist": "39 sange",
    "title": "Kunstnerliste: Dans & L\xe6r",
    "album_art_uri": "http://varnish01.music.aspiro.com/im/im?h=160&w=240&"
    "rows=1&cols=2&artimg&uuid=133fe1ff-1f16-4300-b440-"
    "fb80573f19ce",
    "uri": "x-rincon-cpcontainer:000d006cplaylistid_133fe1ff-1f16-4300-b440"
    "-fb80573f19ce",
    "parent_id": "00020064playlistsearch:Dans &",
    "item_type": "albumList",
    "extended_id": "000d006cplaylistid_133fe1ff-1f16-4300-b440-fb80573f19ce",
    "item_id": "playlistid_133fe1ff-1f16-4300-b440-fb80573f19ce",
    "service_id": 20,
    "artist_id": "artistid_0",
    "can_play": True,
    "can_enumerate": True,
}


class FakeMusicService:
    """A fake music service."""

    def __init__(self, username):
        self.description = "SA_RINCON5127_{}".format(username)
        self.service_id = 20

    @staticmethod
    def id_to_extended_id(item_id, item_class):
        """ID to extended ID method."""
        id_prefix = {
            MSTrack: "00030020",
            MSAlbum: "0004002c",
            MSArtist: "10050024",
            MSAlbumList: "000d006c",
            MSPlaylist: "0006006c",
            MSArtistTracklist: "100f006c",
            MSFavorites: None,  # This one is unknown
            MSCollection: None,  # This one is unknown
        }
        out = id_prefix[item_class]
        if out:
            out += item_id
        return out

    @staticmethod
    def form_uri(item_content, item_class):
        """Form the URI."""
        mime_type_to_extension = {"audio/aac": "mp4"}
        uris = {
            MSTrack: "x-sonos-http:{item_id}.{extension}?sid={service_id}&" "flags=32",
            MSAlbum: "x-rincon-cpcontainer:{extended_id}",
            MSAlbumList: "x-rincon-cpcontainer:{extended_id}",
            MSPlaylist: "x-rincon-cpcontainer:{extended_id}",
            MSArtistTracklist: "x-rincon-cpcontainer:{extended_id}",
        }
        extension = None
        if "mime_type" in item_content:
            extension = mime_type_to_extension[item_content["mime_type"]]
        out = uris.get(item_class)
        if out:
            # pylint: disable=star-args
            out = out.format(extension=extension, **item_content)
        return out


FAKE_MUSIC_SERVICE = FakeMusicService("4542255535")


def getter_attributes_test(name, from_xml, from_dict, result):
    """Test if the getters return the right value."""
    assert getattr(from_xml, name) == result
    assert getattr(from_dict, name) == result


def common_tests(class_, xml_, dict_, parent_id, helpers):
    """Common tests for the MS classes."""
    xml_content = XML.fromstring(xml_.encode("utf8"))

    # MusicServiceItem.from_xml and MusicServiceItem.to_dict
    item_from_xml = class_.from_xml(xml_content, FAKE_MUSIC_SERVICE, parent_id)
    assert item_from_xml.to_dict == dict_

    # MusicServiceItem.from_dict and MusicServiceItem.to_dict
    item_from_dict = class_.from_dict(dict_)
    assert item_from_dict.to_dict == dict_

    # MusicServiceItem.didl_metadata
    # NOTE! These tests is reliant on the attributes being put in the same
    # order by ElementTree and so it might fail if that changes
    if item_from_xml.can_play:
        dict_encoded = {}
        for key, value in dict_.items():
            try:
                is_str = isinstance(value, unicode)
            except NameError:
                is_str = isinstance(value, str)
            if is_str:
                dict_encoded[key] = (
                    escape(value).encode("ascii", "xmlcharrefreplace").decode("ascii")
                )

            else:
                dict_encoded[key] = value
        didl = DIDL_TEMPLATE.format(item_class=class_.item_class, **dict_encoded)

        assert helpers.compare_xml(item_from_xml.didl_metadata, XML.fromstring(didl))
        assert helpers.compare_xml(item_from_dict.didl_metadata, XML.fromstring(didl))
    else:
        with pytest.raises(DIDLMetadataError):
            # pylint: disable=pointless-statement
            item_from_xml.didl_metadata

    # Text attributes with mandatory content
    for name in ["item_id", "extended_id", "title", "service_id"]:
        getter_attributes_test(name, item_from_xml, item_from_dict, dict_[name])
    # Text attributes with voluntary content
    for name in ["parent_id", "album_art_uri"]:
        getter_attributes_test(name, item_from_xml, item_from_dict, dict_.get(name))
    # Boolean attribute
    getter_attributes_test(
        "can_play", item_from_xml, item_from_dict, bool(dict_.get("can_play"))
    )
    return item_from_xml, item_from_dict


def test_ms_track_search(helpers):
    """Test the MSTrack item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSTrack,
        MS_TRACK_SEARCH_XML,
        MS_TRACK_SEARCH_DICT,
        "00020064tracksearch:pilgrim",
        helpers,
    )
    getter_attributes_test(
        "artist", item_from_xml, item_from_dict, MS_TRACK_SEARCH_DICT.get("artist")
    )
    getter_attributes_test(
        "uri", item_from_xml, item_from_dict, MS_TRACK_SEARCH_DICT["uri"]
    )


def test_ms_album_search(helpers):
    """Test the MSAlbum item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSAlbum,
        MS_ALBUM_SEARCH_XML,
        MS_ALBUM_SEARCH_DICT,
        "00020064albumsearch:de unge",
        helpers,
    )
    getter_attributes_test(
        "artist", item_from_xml, item_from_dict, MS_ALBUM_SEARCH_DICT.get("artist")
    )
    getter_attributes_test(
        "uri", item_from_xml, item_from_dict, MS_ALBUM_SEARCH_DICT["uri"]
    )


def test_ms_artist_search(helpers):
    """Test the MSAlbum item when instantiated from a search."""
    common_tests(
        MSArtist,
        MS_ARTIST_SEARCH_XML,
        MS_ARTIST_SEARCH_DICT,
        "00020064artistsearch:Fritjof",
        helpers,
    )


def test_ms_playlist_search(helpers):
    """Test the MSAlbum item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSAlbumList,
        MS_PLAYLIST_SEARCH_XML,
        MS_PLAYLIST_SEARCH_DICT,
        "00020064playlistsearch:Dans &",
        helpers,
    )
    getter_attributes_test(
        "uri", item_from_xml, item_from_dict, MS_PLAYLIST_SEARCH_DICT["uri"]
    )



from unittest import mock
import pytest

from pysonos import SoCo
from pysonos.exceptions import SoCoUPnPException

IP_ADDR = "192.168.1.101"


@pytest.fixture()
def moco():
    """A mock soco with fake services and hardcoded is_coordinator.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        "AVTransport",
        "RenderingControl",
        "DeviceProperties",
        "ContentDirectory",
        "ZoneGroupTopology",
    )
    patchers = [mock.patch("pysonos.core.{}".format(service)) for service in services]
    for patch in patchers:
        patch.start()
    with mock.patch(
        "pysonos.SoCo.is_coordinator", new_callable=mock.PropertyMock
    ) as is_coord:
        is_coord = True
        yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


class TestMusicLibrary:
    def test_search_track_no_result(self, moco):
        moco.contentDirectory.reset_mock()
        # Browse returns an exception if the artist can't be found
        # <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
        #  s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
        #   <s:Body>
        #     <s:Fault>
        #       <faultcode>s:Client</faultcode>
        #       <faultstring>UPnPError</faultstring>
        #       <detail>
        #         <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
        #           <errorCode>701</errorCode>
        #         </UPnPError>
        #       </detail>
        #     </s:Fault>
        #   </s:Body>
        # </s:Envelope>
        moco.contentDirectory.Browse.side_effect = SoCoUPnPException(
            "No such object", "701", "error XML"
        )

        result = moco.music_library.search_track("artist")

        assert len(result) == 0

        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "A:ALBUMARTIST/artist/"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 100000),
                ("SortCriteria", ""),
            ]
        )

    def test_search_track_no_artist_album_track(self, moco):
        moco.contentDirectory.reset_mock()
        # Browse returns an empty result set if artist and album and
        # track cannot be found
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            "NumberReturned": "0",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"></DIDL-Lite>'
            ),
            "TotalMatches": "0",
            "UpdateID": "0",
        }

        result = moco.music_library.search_track("artist", "album", "track")

        assert len(result) == 0

        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "A:ALBUMARTIST/artist/album:track"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 100000),
                ("SortCriteria", ""),
            ]
        )

    def test_search_track_artist_albums(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            "NumberReturned": "2",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:ns0="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
                'xmlns:ns2="urn:schemas-upnp-org:metadata-1-0/upnp/">'
                '<container id="A:ALBUMARTIST/The%20Artist/First%20Album" '
                'parentID="A:ALBUMARTIST/The%20Artist" restricted="true">'
                "<dc:title>First Album</dc:title>"
                "<ns2:class>object.container.album.musicAlbum</ns2:class>"
                '<res protocolInfo="x-rincon-playlist:*:*:*">'
                "x-rincon-playlist:RINCON_000123456789001400#A:ALBUMARTIST/"
                "The%20Artist/First%20Album</res>"
                "<dc:creator>The Artist</dc:creator>"
                "<ns2:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist"
                "%2fFirst%2520Album%2ftrack2.mp3&amp;v=432</ns2:albumArtURI>"
                "</container>"
                '<container id="A:ALBUMARTIST/The%20Artist/Second%20Album" '
                'parentID="A:ALBUMARTIST/The%20Artist" restricted="true">'
                "<dc:title>Second Album</dc:title>"
                "<ns2:class>object.container.album.musicAlbum</ns2:class>"
                '<res protocolInfo="x-rincon-playlist:*:*:*">x-rincon-playlist:'
                "RINCON_000123456789001400#A:ALBUMARTIST/The%20Artist/Second%20Album"
                "</res>"
                "<dc:creator>The Artist</dc:creator>"
                "<ns2:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist"
                "%2fSecond%2520Album%2ftrack2.mp3&amp;v=432</ns2:albumArtURI>"
                "</container>"
                "</DIDL-Lite>"
            ),
            "TotalMatches": "2",
            "UpdateID": "0",
        }
        results = moco.music_library.search_track("The Artist")

        assert results.number_returned == 2

        album = results[0]
        assert album.title == "First Album"
        assert album.item_id == "A:ALBUMARTIST/The%20Artist/First%20Album"
        album = results[1]
        assert album.title == "Second Album"
        assert album.item_id == "A:ALBUMARTIST/The%20Artist/Second%20Album"

        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "A:ALBUMARTIST/The%20Artist/"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 100000),
                ("SortCriteria", ""),
            ]
        )

    def test_search_track_artist_album_tracks(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            "NumberReturned": "3",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:ns0="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
                'xmlns:ns1="urn:schemas-upnp-org:metadata-1-0/upnp/">'
                '<item id="S://server/The%20Artist/The%20Album/'
                '03%20-%20Track%20Title%201.mp3" '
                'parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true">'
                '<res protocolInfo="x-file-cifs:*:audio/mpeg:*">x-file-cifs://server/'
                "The%20Artist/The%20Album/03%20-%20Track%20Title%201.mp3</res>"
                "<ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist"
                "%2fThe%2520Album%2f03%2520-%2520Track%2520Title%25201.mp3&amp;v=432"
                "</ns1:albumArtURI>"
                "<dc:title>Track Title 1</dc:title>"
                "<ns1:class>object.item.audioItem.musicTrack</ns1:class>"
                "<dc:creator>The Artist</dc:creator>"
                "<ns1:album>The Album</ns1:album>"
                "<ns1:originalTrackNumber>3</ns1:originalTrackNumber>"
                "</item>"
                '<item id="S://server/The%20Artist/The%20Album/'
                '04%20-%20Track%20Title%202.m4a" '
                'parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true">'
                '<res protocolInfo="x-file-cifs:*:audio/mp4:*">x-file-cifs://server/'
                "The%20Artist/The%20Album/04%20-%20Track%20Title%202.m4a</res>"
                "<ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520"
                "Artist%2fThe%2520Album%2f04%2520-%2520Track%2520Title%25202.m4a"
                "&amp;v=432</ns1:albumArtURI>"
                "<dc:title>Track Title 2</dc:title>"
                "<ns1:class>object.item.audioItem.musicTrack</ns1:class>"
                "<dc:creator>The Artist</dc:creator>"
                "<ns1:album>The Album</ns1:album>"
                "<ns1:originalTrackNumber>4</ns1:originalTrackNumber>"
                "</item>"
                '<item id="S://server/The%20Artist/The%20Album/'
                '05%20-%20Track%20Title%203.mp3" '
                'parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true">'
                '<res protocolInfo="x-file-cifs:*:audio/mpeg:*">'
                "x-file-cifs://server/The%20Artist/The%20Album/"
                "05%20-%20Track%20Title%203.mp3</res>"
                "<ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520"
                "Artist%2fThe%2520Album%2f05%2520-%2520Track%2520Title%25203.mp3"
                "&amp;v=432</ns1:albumArtURI>"
                "<dc:title>Track Title 3</dc:title>"
                "<ns1:class>object.item.audioItem.musicTrack</ns1:class>"
                "<dc:creator>The Artist</dc:creator>"
                "<ns1:album>The Album</ns1:album>"
                "<ns1:originalTrackNumber>5</ns1:originalTrackNumber>"
                "</item>"
                "</DIDL-Lite>"
            ),
            "TotalMatches": "3",
            "UpdateID": "0",
        }

        results = moco.music_library.search_track("The Artist", "The Album")

        assert len(results) == 3

        track = results[0]
        assert track.title == "Track Title 1"
        assert (
            track.item_id
            == "S://server/The%20Artist/The%20Album/03%20-%20Track%20Title%201.mp3"
        )
        track = results[1]
        assert track.title == "Track Title 2"
        assert (
            track.item_id
            == "S://server/The%20Artist/The%20Album/04%20-%20Track%20Title%202.m4a"
        )
        track = results[2]
        assert track.title == "Track Title 3"
        assert (
            track.item_id
            == "S://server/The%20Artist/The%20Album/05%20-%20Track%20Title%203.mp3"
        )

        moco.contentDirectory.Browse.assert_called_once_with(
            [
                ("ObjectID", "A:ALBUMARTIST/The%20Artist/The%20Album"),
                ("BrowseFlag", "BrowseDirectChildren"),
                ("Filter", "*"),
                ("StartingIndex", 0),
                ("RequestedCount", 100000),
                ("SortCriteria", ""),
            ]
        )

    def test_soco_library_updating(self, moco):
        moco.contentDirectory.GetShareIndexInProgress.return_value = {"IsIndexing": "0"}
        assert not moco.music_library.library_updating
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.GetShareIndexInProgress.return_value = {"IsIndexing": "1"}
        assert moco.music_library.library_updating

    def test_soco_start_library_update(self, moco):
        moco.contentDirectory.RefreshShareIndex.return_value = True
        assert moco.music_library.start_library_update()
        moco.contentDirectory.RefreshShareIndex.assert_called_with(
            [
                ("AlbumArtistDisplayOption", ""),
            ]
        )

    def test_soco_list_library_shares(self, moco):
        # Tests with 2, 1 and 0 library shares
        moco.contentDirectory.Browse.return_value = {
            "TotalMatches": "2",
            "NumberReturned": "2",
            "UpdateID": "0",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                '<container id="S://share_host/music_01/Music/Lossless" '
                'parentID="S:" restricted="true"><dc:title>//share_host/music_01/Music/'
                "Lossless"
                "</dc:title><upnp:class>object.container</upnp:class><res "
                'protocolInfo="x-rincon-playlist:*:*:*">'
                "x-rincon-playlist:RINCON_XXXXXXXXXXXXX1400"
                "#S://share_host/music_01/Music/Lossless</res></container>"
                '<container id="S://share_host_2/music_01" parentID="S:" '
                'restricted="true">'
                "<dc:title>//share_host_2/music_01</dc:title>"
                "<upnp:class>object.container"
                '</upnp:class><res protocolInfo="x-rincon-playlist:*:*:*">'
                "x-rincon-playlist"
                ":RINCON_XXXXXXXXXXXXX1400#S://share_host_2/music_01</res>"
                "</container></DIDL-Lite>"
            ),
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 2
        assert "//share_host/music_01/Music/Lossless" in results
        assert "//share_host_2/music_01" in results

        moco.contentDirectory.Browse.return_value = {
            "TotalMatches": "1",
            "NumberReturned": "1",
            "UpdateID": "0",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                '<container id="S://share_host/music_01/Music/Lossless" parentID="S:" '
                'restricted="true">'
                "<dc:title>//share_host/music_01/Music/Lossless</dc:title>"
                "<upnp:class>object.container</upnp:class>"
                '<res protocolInfo="x-rincon-playlist:*:*:*">'
                "x-rincon-playlist:RINCON_XXXXXXXXXXXXXX400#"
                "S://share_host/music_01/Music/Lossless</res></container></DIDL-Lite>"
            ),
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 1
        assert "//share_host/music_01/Music/Lossless" in results

        moco.contentDirectory.Browse.return_value = {
            "TotalMatches": "0",
            "NumberReturned": "0",
            "UpdateID": "0",
            "Result": (
                '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"/>'
            ),
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 0

    def test_soco_delete_library_share(self, moco):
        share = "//host/share"
        moco.music_library.delete_library_share(share)
        moco.contentDirectory.DestroyObject.assert_called_once_with(
            [("ObjectID", "S:" + share)]
        )



# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance,no-else-continue

"""This module contains classes relating to Third Party music services."""


import logging
import weakref

import requests

from .. import discovery
from ..xml import XML

log = logging.getLogger(__name__)  # pylint: disable=C0103


# pylint: disable=too-many-instance-attributes
class Account:

    """An account for a Music Service.

    Each service may have more than one account: see the `Sonos release notes
    for version 5-2 <http://www.sonos.com/en-gb/software/release/5-2>`_
    """

    _all_accounts = weakref.WeakValueDictionary()

    def __init__(self):
        super().__init__()
        #: str: A unique identifier for the music service to which this
        #: account relates, eg ``'2311'`` for Spotify.
        self.service_type = ""
        #: str: A unique identifier for this account
        self.serial_number = ""
        #: str: The account's nickname
        self.nickname = ""
        #: bool: `True` if this account has been deleted
        self.deleted = False
        #: str: The username used for logging into the music service
        self.username = ""
        #: str: Metadata for the account
        self.metadata = ""
        #: str: Used for OpenAuth id for some services
        self.oa_device_id = ""
        #: str: Used for OpenAuthid for some services
        self.key = ""

    def __repr__(self):
        return "<{} '{}:{}:{}' at {}>".format(
            self.__class__.__name__,
            self.serial_number,
            self.service_type,
            self.nickname,
            hex(id(self)),
        )

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def _get_account_xml(soco):
        """Fetch the account data from a Sonos device.

        Args:
            soco (SoCo): a SoCo instance to query. If soco is `None`, a
                random device will be used.

        Returns:
            str: a byte string containing the account data xml
        """
        # It is likely that the same information is available over UPnP as well
        # via a call to
        # systemProperties.GetStringX([('VariableName','R_SvcAccounts')]))
        # This returns an encrypted string, and, so far, we cannot decrypt it
        device = soco or discovery.any_soco()
        log.debug("Fetching account data from %s", device)
        settings_url = "http://{}:1400/status/accounts".format(device.ip_address)
        result = requests.get(settings_url).content
        log.debug("Account data: %s", result)
        return result

    @classmethod
    def get_accounts(cls, soco=None):
        """Get all accounts known to the Sonos system.

        Args:
            soco (`SoCo`, optional): a `SoCo` instance to query. If `None`, a
                random instance is used. Defaults to `None`.

        Returns:
            dict: A dict containing account instances. Each key is the
            account's serial number, and each value is the related Account
            instance. Accounts which have been marked as deleted are excluded.

        Note:
            Any existing Account instance will have its attributes updated
            to those currently stored on the Sonos system.
        """

        root = XML.fromstring(cls._get_account_xml(soco))
        # _get_account_xml returns an ElementTree element like this:

        # <ZPSupportInfo type="User">
        #   <Accounts
        #   LastUpdateDevice="RINCON_000XXXXXXXX400"
        #   Version="8" NextSerialNum="5">
        #     <Account Type="2311" SerialNum="1">
        #         <UN>12345678</UN>
        #         <MD>1</MD>
        #         <NN></NN>
        #         <OADevID></OADevID>
        #         <Key></Key>
        #     </Account>
        #     <Account Type="41735" SerialNum="3" Deleted="1">
        #         <UN></UN>
        #         <MD>1</MD>
        #         <NN>Nickname</NN>
        #         <OADevID></OADevID>
        #         <Key></Key>
        #     </Account>
        # ...
        #   <Accounts />

        xml_accounts = root.findall(".//Account")
        result = {}
        for xml_account in xml_accounts:
            serial_number = xml_account.get("SerialNum")
            is_deleted = xml_account.get("Deleted") == "1"
            # cls._all_accounts is a weakvaluedict keyed by serial number.
            # We use it as a database to store details of the accounts we
            # know about. We need to update it with info obtained from the
            # XML just obtained, so (1) check to see if we already have an
            # entry in cls._all_accounts for the account we have found in
            # XML; (2) if so, delete it if the XML says it has been deleted;
            # and (3) if not, create an entry for it
            if cls._all_accounts.get(serial_number):
                # We have an existing entry in our database. Do we need to
                # delete it?
                if is_deleted:
                    # Yes, so delete it and move to the next XML account
                    del cls._all_accounts[serial_number]
                    continue
                else:
                    # No, so load up its details, ready to update them
                    account = cls._all_accounts.get(serial_number)
            else:
                # We have no existing entry for this account
                if is_deleted:
                    # but it is marked as deleted, so we don't need one
                    continue
                # If it is not marked as deleted, we need to create an entry
                account = Account()
                account.serial_number = serial_number
                cls._all_accounts[serial_number] = account

            # Now, update the entry in our database with the details from XML
            account.service_type = xml_account.get("Type")
            account.deleted = is_deleted
            account.username = xml_account.findtext("UN")
            # Not sure what 'MD' stands for.  Metadata? May Delete?
            account.metadata = xml_account.findtext("MD")
            account.nickname = xml_account.findtext("NN")
            account.oa_device_id = xml_account.findtext("OADevID")
            account.key = xml_account.findtext("Key")
            result[serial_number] = account
        # There is always a TuneIn account, but it is handled separately
        #  by Sonos, and does not appear in the xml account data. We
        # need to add it ourselves.
        tunein = Account()
        tunein.service_type = "65031"  # Is this always the case?
        tunein.deleted = False
        tunein.username = ""
        tunein.metadata = ""
        tunein.nickname = ""
        tunein.oa_device_id = ""
        tunein.key = ""
        tunein.serial_number = "0"
        result["0"] = tunein

        return result

    @classmethod
    def get_accounts_for_service(cls, service_type):
        """Get a list of accounts for a given music service.

        Args:
            service_type (str): The service_type to use.

        Returns:
            list: A list of `Account` instances.
        """
        return [
            a for a in cls.get_accounts().values() if a.service_type == service_type
        ]



"""Tests for the utils module."""


from pysonos.utils import deprecated


# Deprecation decorator


def test_deprecation(recwarn):
    @deprecated("0.7")
    def dummy(args):
        """My docs."""
        pass

    @deprecated("0.8", "better_function", "0.12")
    def dummy2(args):
        """My docs."""
        pass

    assert dummy.__doc__ == "My docs.\n\n  .. deprecated:: 0.7\n"
    assert (
        dummy2.__doc__ == "My docs.\n\n  .. deprecated:: 0.8\n\n"
        "     Will be removed in version 0.12.\n"
        "     Use `better_function` instead."
    )
    dummy(3)
    w = recwarn.pop()
    assert str(w.message) == "Call to deprecated function dummy."
    dummy2(4)
    w = recwarn.pop()
    assert (
        str(w.message) == "Call to deprecated function dummy2. Will be "
        "removed in version 0.12. Use "
        "better_function instead."
    )
    assert w.filename
    assert w.lineno



# pylint: disable-msg=R0904

"""This file contains the classes used to perform unit tests on the methods in
the SoCo class.

PLEASE TAKE NOTE: All of these unit tests are designed to run on a sonos
system without interfering with normal service. This means that they will not
raise the volume or leave the player in another state than it started in. They
have been made this way, since sonos is developed by volunteers, that in all
likelihood does not have a dedicated test system, so the tests must be able to
run on an ordinary system without annoying the neighboors and it should return
to its original state because those same developers will likely want to listen
to music while coding, without having it interrupted at every unit test.
PLEASE RESPECT THIS.
"""

import time
import unittest

import pytest

import pysonos

SOCO = None
pytestmark = pytest.mark.integration


class SoCoUnitTestInitError(Exception):
    """Exception for incomplete unit test initialization."""

    def __init__(self, message):
        Exception.__init__(self, message)


def init(**kwargs):
    """Initialize variables for the unittests that are only known at run
    time."""
    global SOCO  # pylint: disable-msg=W0603
    SOCO = pysonos.SoCo(kwargs["ip"])

    if len(SOCO.get_queue()) == 0:
        raise SoCoUnitTestInitError(
            "Unit tests on the SoCo class must be run "
            "with at least 1 item in the playlist"
        )

    transport_info = SOCO.get_current_transport_info()
    if transport_info["current_transport_state"] != "PLAYING":
        raise SoCoUnitTestInitError(
            "Unit tests on the SoCo class must be run " "with the sonos unit playing"
        )


def get_state():
    """Utility function to get the entire playing state before the unit tests
    starts to change it."""
    state = {
        "queue": SOCO.get_queue(0, 1000),
        "current_track_info": SOCO.get_current_track_info(),
    }
    return state


def set_state(state):
    """Utility function to set the entire state.

    Used to reset the unit after the unit tests have changed it
    """
    SOCO.stop()
    SOCO.clear_queue()
    for track in state["queue"]:
        SOCO.add_to_queue(track["uri"])
    SOCO.play_from_queue(int(state["current_track_info"]["playlist_position"]) - 1)
    SOCO.seek(state["current_track_info"]["position"])
    SOCO.play()


def wait(interval=0.1):
    """Convinience function to adjust sleep interval for all tests."""
    time.sleep(interval)


# Test return strings that are used a lot
NOT_TRUE = "The method did not return True"
NOT_EXP = "The method did not return the expected value"
NOT_TYPE = "The return value of the method did not have the expected type: {}"
NOT_IN_RANGE = "The returned value is not in the expected range"


# functions for running via pytest
def setup_module(module):
    ip = pytest.config.option.IP
    if ip is None:
        pytest.fail("No ip address specified. Use the --ip option.")
    init(ip=ip)
    state = get_state()
    module.state = state


def teardown_module(module):
    state = module.state
    set_state(state)


class Volume(unittest.TestCase):
    """Unit tests for the volume method."""

    def setUp(self):  # pylint: disable-msg=C0103
        self.valid_values = range(101)

    def test_get_and_set(self):
        """Tests if the set functionlity works when given valid arguments."""
        old = SOCO.volume
        self.assertIn(old, self.valid_values, NOT_IN_RANGE)
        if old == self.valid_values[0]:
            new = old + 1
        else:
            new = old - 1
        SOCO.volume = new
        wait()
        self.assertEqual(SOCO.volume, new, NOT_EXP)
        SOCO.volume = old
        wait()

    def test_invalid_arguments(self):
        """Tests if the set functionality coerces into range when given
        integers outside of allowed range."""
        old = SOCO.volume
        # NOTE We don't test coerce from too large values, since that would
        # put the unit at full volume
        SOCO.volume = self.valid_values[0] - 1
        wait()
        self.assertEqual(SOCO.volume, 0, NOT_EXP)
        SOCO.volume = old
        wait()

    def test_set_0(self):
        """Tests whether the volume can be set to 0. Regression test for:
        https://github.com/rahims/SoCo/issues/29
        """
        old = SOCO.volume
        SOCO.volume = 0
        wait()
        self.assertEqual(SOCO.volume, 0, NOT_EXP)
        SOCO.volume = old
        wait()


class Bass(unittest.TestCase):
    """Unit tests for the bass method.

    This class implements a full boundary value test.
    """

    def setUp(self):  # pylint: disable-msg=C0103
        self.valid_values = range(-10, 11)

    def test_get_and_set(self):
        """Tests if the set functionlity works when given valid arguments."""
        old = SOCO.bass
        self.assertIn(old, self.valid_values, NOT_IN_RANGE)
        # Values on the boundaries of the valid equivalence partition
        for value in [self.valid_values[0], self.valid_values[-1]]:
            SOCO.bass = value
            wait()
            self.assertEqual(SOCO.bass, value, NOT_EXP)
        SOCO.bass = old
        wait()

    def test_invalid_arguments(self):
        """Tests if the set functionality produces the expected "coerce in
        range" functionality when given a value outside of its range."""
        old = SOCO.bass
        # Values on the boundaries of the two invalid equivalence partitions
        SOCO.bass = self.valid_values[0] - 1
        wait()
        self.assertEqual(SOCO.bass, self.valid_values[0], NOT_EXP)
        SOCO.bass = self.valid_values[-1] + 1
        wait()
        self.assertEqual(SOCO.bass, self.valid_values[-1], NOT_EXP)
        SOCO.bass = old
        wait()


class Treble(unittest.TestCase):
    """Unit tests for the treble method This class implements a full boundary
    value test."""

    def setUp(self):  # pylint: disable-msg=C0103
        self.valid_values = range(-10, 11)

    def test_get_and_set(self):
        """Tests if the set functionlity works when given valid arguments."""
        old = SOCO.treble
        self.assertIn(old, self.valid_values, NOT_IN_RANGE)
        # Values on the boundaries of the valid equivalence partition
        for value in [self.valid_values[0], self.valid_values[-1]]:
            SOCO.treble = value
            wait()
            self.assertEqual(SOCO.treble, value, NOT_EXP)
        SOCO.treble = old
        wait()

    def test_invalid_arguments(self):
        """Tests if the set functionality produces the expected "coerce in
        range" functionality when given a value outside its range."""
        old = SOCO.treble
        # Values on the boundaries of the two invalid equivalence partitions
        SOCO.treble = self.valid_values[0] - 1
        wait()
        self.assertEqual(SOCO.treble, self.valid_values[0], NOT_EXP)
        SOCO.treble = self.valid_values[-1] + 1
        wait()
        self.assertEqual(SOCO.treble, self.valid_values[-1], NOT_EXP)
        SOCO.treble = old
        wait()


class GetCurrentTrackInfo(unittest.TestCase):
    """Unit test for the get_current_track_info method."""

    def setUp(self):  # pylint: disable-msg=C0103
        # The value in this list must be kept up to date with the values in
        # the test_get doc string
        self.info_keys = sorted(
            [
                "album",
                "artist",
                "title",
                "uri",
                "playlist_position",
                "duration",
                "album_art",
                "position",
            ]
        )

    def test_get(self):
        """Test is the return value is a dictinary and contains the following
        keys: album, artist, title, uri, playlist_position, duration,
        album_art and position
        """
        info = SOCO.get_current_track_info()
        self.assertIsInstance(info, dict, "Returned info is not a dict")
        self.assertEqual(
            sorted(info.keys()), self.info_keys, "Info does not contain the proper keys"
        )


class AddToQueue(unittest.TestCase):
    """Unit test for the add_to_queue method."""

    def test(self):
        """Gets the current queue, adds the last item of the current queue and
        then compares the length of the old queue with the new and checks that
        the last two elements are identical."""
        state = get_state()
        SOCO.pause()
        old_queue = SOCO.get_queue(0, 1000)
        # Add new element and check
        self.assertEqual(
            SOCO.add_to_queue(old_queue[-1]["uri"]), len(old_queue) + 1, ""
        )
        wait()
        new_queue = SOCO.get_queue()
        self.assertEqual(len(new_queue) - 1, len(old_queue))
        self.assertEqual(new_queue[-1], new_queue[-2])
        # Clean up
        set_state(state)
        wait()


class GetQueue(unittest.TestCase):
    """Unit test for the get_queue method."""

    def setUp(self):  # pylint: disable-msg=C0103
        # The values in this list must be kept up to date with the values in
        # the test_get doc string
        self.qeueu_element_keys = sorted(
            ["album", "artist", "uri", "album_art", "title"]
        )

    def test_get(self):
        """Tests is return value is a list of dictionaries and if each of
        the dictionaries contain the keys: album, artist, uri, album_art and
        title
        """
        queue = SOCO.get_queue()
        self.assertIsInstance(queue, list, NOT_TYPE.format("list"))
        for item in queue:
            self.assertIsInstance(item, dict, "Item in queue is not a dictionary")
            self.assertEqual(
                sorted(item.keys()),
                self.qeueu_element_keys,
                "The keys in the queue element dict are not the "
                "expected ones: {}".format(self.qeueu_element_keys),
            )


class GetCurrentTransportInfo(unittest.TestCase):
    """Unit test for the get_current_transport_info method."""

    def setUp(self):  # pylint: disable-msg=C0103
        # The values in this list must be kept up to date with the values in
        # the test doc string
        self.transport_info_keys = sorted(
            [
                "current_transport_status",
                "current_transport_state",
                "current_transport_speed",
            ]
        )

    def test(self):
        """Tests if the return value is a dictionary that contains the keys:
        current_transport_status, current_transport_state,
        current_transport_speed
        and that values have been found for all keys, i.e. they are not None
        """
        transport_info = SOCO.get_current_transport_info()
        self.assertIsInstance(transport_info, dict, NOT_TYPE.format("dict"))
        self.assertEqual(
            self.transport_info_keys,
            sorted(transport_info.keys()),
            "The keys in the speaker info dict are not the "
            "expected ones: {}".format(self.transport_info_keys),
        )
        for key, value in transport_info.items():
            self.assertIsNotNone(
                value,
                'The value for the key "{}" is None '
                "which indicate that no value was found for "
                "it".format(key),
            )


class GetSpeakerInfo(unittest.TestCase):
    """Unit test for the get_speaker_info method."""

    def setUp(self):  # pylint: disable-msg=C0103
        # The values in this list must be kept up to date with the values in
        # the test doc string
        self.info_keys = sorted(
            [
                "zone_name",
                "player_icon",
                "uid",
                "serial_number",
                "software_version",
                "hardware_version",
                "mac_address",
                "model_name",
                "model_number",
                "display_version",
            ]
        )

    def test(self):
        """Tests if the return value is a dictionary that contains the keys:
        zone_name, player_icon, uid, serial_number, software_version,
        hardware_version, mac_address, model_name, model_number, display_version
        and that values have been found for all keys, i.e. they are not None
        """
        speaker_info = SOCO.get_speaker_info()
        self.assertIsInstance(speaker_info, dict, NOT_TYPE.format("dict"))
        self.assertEqual(
            self.info_keys,
            sorted(speaker_info.keys()),
            "The "
            "keys in speaker info are not the expected ones: {}"
            "".format(self.info_keys),
        )
        for key, value in speaker_info.items():
            self.assertIsNotNone(
                value,
                'The value for the key "{}" is None '
                "which indicate that no value was found for "
                "it".format(key),
            )


# class GetSpeakersIp(unittest.TestCase):
# """ Unit tests for the get_speakers_ip method """

# TODO: Awaits https://github.com/rahims/SoCo/issues/26

# def test(self):
# print SOCO.get_speakers_ip()


class Pause(unittest.TestCase):
    """Unittest for the pause method."""

    def test(self):
        """Tests if the pause method works."""
        SOCO.pause()
        wait(1)
        new = SOCO.get_current_transport_info()["current_transport_state"]
        self.assertEqual(
            new, "PAUSED_PLAYBACK", "State after pause is not " '"PAUSED_PLAYBACK"'
        )
        SOCO.play()
        wait(1)


class Stop(unittest.TestCase):
    """Unittest for the stop method."""

    def test(self):
        """Tests if the stop method works."""
        state = get_state()
        SOCO.stop()
        wait(1)
        new = SOCO.get_current_transport_info()["current_transport_state"]
        self.assertEqual(new, "STOPPED", 'State after stop is not "STOPPED"')
        set_state(state)  # Reset unit the way it was before the test
        wait(1)


class Play(unittest.TestCase):
    """Unit test for the play method."""

    def test(self):
        """Tests if the play method works."""
        SOCO.pause()
        wait(1)
        on_pause = SOCO.get_current_transport_info()["current_transport_state"]
        self.assertEqual(
            on_pause, "PAUSED_PLAYBACK", "State after pause is " 'not "PAUSED_PLAYBACK"'
        )
        SOCO.play()
        wait(1)
        on_play = SOCO.get_current_transport_info()["current_transport_state"]
        self.assertEqual(
            on_play, "PLAYING", "State after play is not " '"PAUSED_PLAYBACK"'
        )


class Mute(unittest.TestCase):
    """Unit test for the mute method."""

    def test(self):
        """Tests of the mute method works."""
        old = SOCO.mute
        self.assertEqual(
            old, 0, "The unit should not be muted when running " "the unit tests"
        )
        SOCO.mute = True
        wait()
        new = SOCO.mute
        self.assertEqual(new, 1, "The unit did not successfully mute")
        SOCO.mute = False
        wait()


class RemoveFromQueue(unittest.TestCase):
    """Unit test for the remove_from_queue method."""

    def test(self):
        """Tests if the remove_from_queue method works."""
        old_queue = SOCO.get_queue()
        track_to_remove = old_queue[-1]
        SOCO.remove_from_queue(len(old_queue))
        wait()
        new_queue = SOCO.get_queue()
        self.assertNotEqual(
            old_queue,
            new_queue,
            "No difference between " "queues before and after removing the last item",
        )
        self.assertEqual(
            len(new_queue),
            len(old_queue) - 1,
            "The length of " "queue after removing a track is not length before - " "1",
        )
        # Clean up
        SOCO.add_to_queue(track_to_remove["uri"])
        wait()
        self.assertEqual(old_queue, SOCO.get_queue(), "Clean up unsuccessful")


class Seek(unittest.TestCase):
    """Unit test for the seek method."""

    def test_valid(self):
        """Tests if the seek method works with valid input."""
        original_position = SOCO.get_current_track_info()["position"]
        # Format 1
        SOCO.seek("0:00:00")
        wait()
        position = SOCO.get_current_track_info()["position"]
        self.assertIn(position, ["0:00:00", "0:00:01"])
        # Reset and format 2
        SOCO.seek(original_position)
        SOCO.seek("00:00:00")
        wait()
        position = SOCO.get_current_track_info()["position"]
        self.assertIn(position, ["0:00:00", "0:00:01"])
        # Clean up
        SOCO.seek(original_position)
        wait()

    def test_invald(self):
        """Tests if the seek method properly fails with invalid input."""
        for string in ["invalid_time_string", "5:12", "6", "aa:aa:aa"]:
            with self.assertRaises(ValueError):
                SOCO.seek(string)



"""ShareLink Plugin."""

import re

from ..plugins import SoCoPlugin
from ..music_services import MusicService
from ..exceptions import SoCoException


class ShareClass:
    """Base class for supported services."""

    def __init__(self, soco):
        """Initialize the share object."""
        self.soco = soco

    def canonical_uri(self, uri):
        """Recognize a share link and return its canonical representation.

        Args:
            uri (str): A URI like `https://tidal.com/browse/album/157273956`.

        Returns:
            str: The canonical URI or None if not recognized.
        """
        raise NotImplementedError

    def service_number(self):
        """Return the service number.

        Returns:
            int: A number identifying the supported music service.
        """
        raise NotImplementedError

    @staticmethod
    def magic():
        """Return magic.

        Returns:
            dict: Magic prefix/key/class values for each share type.
        """
        return {
            "album": {
                "prefix": "x-rincon-cpcontainer:1004206c",
                "key": "00040000",
                "class": "object.container.album.musicAlbum",
            },
            "track": {
                "prefix": "",
                "key": "00032020",
                "class": "object.item.audioItem.musicTrack",
            },
            "playlist": {
                "prefix": "x-rincon-cpcontainer:1006206c",
                "key": "1006206c",
                "class": "object.container.playlistContainer",
            },
        }

    def extract(self, uri):
        """Extract the share type and encoded URI from a share link.

        Returns:
            share_type: The shared type, like "album" or "track".
            encoded_uri: An escaped URI with a service-specific format.
        """
        raise NotImplementedError


class SpotifyShare(ShareClass):
    """Spotify share class."""

    def canonical_uri(self, uri):
        match = re.search(r"spotify.*[:/](album|track|playlist)[:/](\w+)", uri)
        if match:
            return "spotify:" + match.group(1) + ":" + match.group(2)

        return None

    def service_number(self):
        spotify_service = MusicService.get_data_for_name("Spotify")
        return spotify_service["ServiceType"]

    def extract(self, uri):
        spotify_uri = self.canonical_uri(uri)
        share_type = spotify_uri.split(":")[1]
        encoded_uri = spotify_uri.replace(":", "%3a")
        return (share_type, encoded_uri)


class TIDALShare(ShareClass):
    """TIDAL share class."""

    def canonical_uri(self, uri):
        match = re.search(r"https://tidal.*[:/](album|track|playlist)[:/]([\w-]+)", uri)
        if match:
            return "tidal:" + match.group(1) + ":" + match.group(2)

        return None

    def service_number(self):
        return 44551

    def extract(self, uri):
        tidal_uri = self.canonical_uri(uri)
        share_type = tidal_uri.split(":")[1]
        encoded_uri = tidal_uri.replace("tidal:", "").replace(":", "%2f")
        return (share_type, encoded_uri)


class ShareLinkPlugin(SoCoPlugin):
    """A SoCo plugin for playing Spotify/Tidal share links."""

    def __init__(self, soco):
        """Initialize the plugin."""
        super().__init__(soco)
        self.services = [SpotifyShare(soco), TIDALShare(soco)]

    @property
    def name(self):
        return "ShareLink Plugin"

    def is_share_link(self, uri):
        """bool: Is the URI for a supported music service."""
        for service in self.services:
            if service.canonical_uri(uri):
                return True

        return False

    def add_share_link_to_queue(self, uri, position=0, as_next=False):
        """Add a Spotify/Tidal/... item to the queue.

        This is similar to soco.add_uri_to_queue() but will work with
        music service share links that do not directly point to sound
        files.

        Args:
            uri (str): A URI like `spotify:album:6wiUBliPe76YAVpNEdidpY`.
            position (int): The index (1-based) at which the URI should be
                added. Default is 0 (add URI at the end of the queue).
            as_next (bool): Whether this URI should be played as the next
                track in shuffle mode. This only works if `play_mode=SHUFFLE`.

        Returns:
            int: The index of the new item in the queue.
        """
        for service in self.services:
            if service.canonical_uri(uri):
                (share_type, encoded_uri) = service.extract(uri)
                magic = service.magic()

                enqueue_uri = magic[share_type]["prefix"] + encoded_uri

                metadata_template = (
                    '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements'
                    '/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata'
                    '-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-'
                    'com:metadata-1-0/" xmlns="urn:schemas-upnp-org:m'
                    'etadata-1-0/DIDL-Lite/"><item id="{item_id}" res'
                    'tricted="true"><upnp:class>{item_class}</upnp:cl'
                    'ass><desc id="cdudn" nameSpace="urn:schemas-rinc'
                    'onnetworks-com:metadata-1-0/">SA_RINCON{sn}_X_#S'
                    "vc{sn}-0-Token</desc></item></DIDL-Lite>"
                )

                metadata = metadata_template.format(
                    item_id=magic[share_type]["key"] + encoded_uri,
                    item_class=magic[share_type]["class"],
                    sn=service.service_number(),
                )

                response = self.soco.avTransport.AddURIToQueue(
                    [
                        ("InstanceID", 0),
                        ("EnqueuedURI", enqueue_uri),
                        ("EnqueuedURIMetaData", metadata),
                        ("DesiredFirstTrackNumberEnqueued", position),
                        ("EnqueueAsNext", int(as_next)),
                    ]
                )
                qnumber = response["FirstTrackNumberEnqueued"]
                return int(qnumber)

        raise SoCoException("Unsupported URI: " + uri)



"""A play local music files example

To use the script:

 * Make sure soco is installed
 * Drop this script into a folder that, besides python files, contains
nothing but music files
 * Choose which player to use and run the script at the command line as such:

play_local_files.py "Living Room"

NOTE: The script has been changed from the earlier version, where the
settings were written directly into the file. They now have to be
given at the command line instead. But, it should only be necessary to
supply the zone name. The local machine IP should be autodetected.

"""



import os
import sys
import time
import socket
from threading import Thread
from random import choice

try:
    # Python 3
    from urllib.parse import quote
    from http.server import SimpleHTTPRequestHandler
    from socketserver import TCPServer

    print("Running as python 3")
except ImportError:
    # Python 2
    from urllib import quote
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from SocketServer import TCPServer

    print("Running as python 2")

from soco.discovery import by_name, discover


class HttpServer(Thread):
    """A simple HTTP Server in its own thread"""

    def __init__(self, port):
        super().__init__()
        self.daemon = True
        handler = SimpleHTTPRequestHandler
        self.httpd = TCPServer(("", port), handler)

    def run(self):
        """Start the server"""
        print("Start HTTP server")
        self.httpd.serve_forever()

    def stop(self):
        """Stop the server"""
        print("Stop HTTP server")
        self.httpd.socket.close()


def add_random_file_from_present_folder(machine_ip, port, zone):
    """Add a random non-py file from this folder and subfolders to soco"""
    # Make a list of music files, right now it is done by collection all files
    # below the current folder whose extension does not start with .py
    # This will probably need to be modded for other pusposes.
    music_files = []
    print("Looking for music files")
    for path, dirs, files in os.walk("."):
        for file_ in files:
            if not os.path.splitext(file_)[1].startswith(".py"):
                music_files.append(os.path.relpath(os.path.join(path, file_)))
                print("Found:", music_files[-1])

    random_file = choice(music_files)
    # urlencode all the path parts (but not the /'s)
    random_file = os.path.join(*[quote(part) for part in os.path.split(random_file)])
    print("\nPlaying random file:", random_file)
    netpath = "http://{}:{}/{}".format(machine_ip, port, random_file)

    number_in_queue = zone.add_uri_to_queue(netpath)
    # play_from_queue indexes are 0-based
    zone.play_from_queue(number_in_queue - 1)


def detect_ip_address():
    """Return the local ip-address"""
    # Rather hackish way to get the local ip-address, recipy from
    # https://stackoverflow.com/a/166589
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip_address = s.getsockname()[0]
    s.close()
    return ip_address


def parse_args():
    """Parse the command line arguments"""
    import argparse

    description = "Play local files with Sonos by running a local web server"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("zone", help="The name of the zone to play from")
    parser.add_argument(
        "--port", default=8000, help="The local machine port to run the webser on"
    )
    parser.add_argument(
        "--ip",
        default=detect_ip_address(),
        help="The local IP address of this machine. By "
        "default it will attempt to autodetect it.",
    )

    return parser.parse_args()


def main():
    # Settings
    args = parse_args()
    print(
        " Will use the following settings:\n"
        " Zone: {args.zone}\n"
        " IP of this machine: {args.ip}\n"
        " Use port: {args.port}".format(args=args)
    )

    # Get the zone
    zone = by_name(args.zone)

    # Check if a zone by the given name was found
    if zone is None:
        zone_names = [zone_.player_name for zone_ in discover()]
        print(
            "No Sonos player named '{}'. Player names are {}".format(
                args.zone, zone_names
            )
        )
        sys.exit(1)

    # Check whether the zone is a coordinator (stand alone zone or
    # master of a group)
    if not zone.is_coordinator:
        print(
            "The zone '{}' is not a group master, and therefore cannot "
            "play music. Please use '{}' in stead".format(
                args.zone, zone.group.coordinator.player_name
            )
        )
        sys.exit(2)

    # Setup and start the http server
    server = HttpServer(args.port)
    server.start()

    # When the http server is setup you can really add your files in
    # any way that is desired. The source code for
    # add_random_file_from_present_folder is just an example, but it may be
    # helpful in figuring out how to format the urls
    try:
        add_random_file_from_present_folder(args.ip, args.port, zone)
        # Remember the http server runs in its own daemonized thread, so it is
        # necessary to keep the main thread alive. So sleep for 3 years.
        time.sleep(10 ** 8)
    except KeyboardInterrupt:
        server.stop()


main()



"""Tests for the cache module."""


from pysonos.cache import Cache, NullCache, TimedCache


def test_instance_creation():
    assert isinstance(Cache(), TimedCache)
    from pysonos import config

    config.CACHE_ENABLED = False
    assert isinstance(Cache(), NullCache)
    config.CACHE_ENABLED = True


def test_cache_put_get():
    """Test putting items into, and getting them from, the cache."""
    from time import sleep

    cache = Cache()
    cache.put("item", "some", kw="args", timeout=3)
    assert not cache.get("some", "otherargs") == "item"
    assert cache.get("some", kw="args") == "item"
    sleep(2)
    assert cache.get("some", kw="args") == "item"
    sleep(2)
    assert not cache.get("some", kw="args") == "item"

    cache.put("item", "some", "args", and_a="keyword", timeout=3)
    assert cache.get("some", "args", and_a="keyword") == "item"
    assert not cache.get("some", "otherargs", and_a="keyword") == "item"


def test_cache_clear_del():
    """Test removal of items and clearing the cache."""
    cache = Cache()
    cache.put("item", "some", kw="args", timeout=2)
    # Check it's there
    assert cache.get("some", kw="args") == "item"
    # delete it
    cache.delete("some", kw="args")
    assert not cache.get("some", kw="args") == "item"
    # put it back
    cache.put("item", "some", kw="args", timeout=3)
    cache.clear()
    assert not cache.get("some", kw="args") == "item"


def test_with_typical_args():
    cache = Cache()
    cache.put(
        "result",
        "SetAVTransportURI",
        [
            ("InstanceID", 1),
            ("CurrentURI", "URI2"),
            ("CurrentURIMetaData", "abcd"),
            ("Unicode", "μИⅠℂ☺ΔЄ💋"),
        ],
        timeout=3,
    )
    assert (
        cache.get(
            "SetAVTransportURI",
            [
                ("InstanceID", 1),
                ("CurrentURI", "URI2"),
                ("CurrentURIMetaData", "abcd"),
                ("Unicode", "μИⅠℂ☺ΔЄ💋"),
            ],
        )
        == "result"
    )


def test_cache_disable():
    cache = Cache()
    assert cache.enabled is True
    cache.enabled = False
    cache.put("item", "args", timeout=3)
    assert cache.get("args") is None
    # Check it's there
    assert cache.get("some", kw="args") is None



# pylint: disable=not-context-manager,useless-object-inheritance

# NOTE: The pylint not-content-manager warning is disabled pending the fix of
# a bug in pylint https://github.com/PyCQA/pylint/issues/782

# NOTE: useless-object-inheritance needed for Python 2.x compatability

"""This module contains the classes underlying SoCo's caching system."""


import threading
from pickle import dumps
from time import time

from . import config


class _BaseCache:

    """An abstract base class for the cache."""

    # pylint: disable=no-self-use, unused-argument

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._cache = {}
        #: `bool`: whether the cache is enabled
        self.enabled = True

    def put(self, item, *args, **kwargs):
        """Put an item into the cache."""
        raise NotImplementedError

    def get(self, *args, **kwargs):
        """Get an item from the cache."""
        raise NotImplementedError

    def delete(self, *args, **kwargs):
        """Delete an item from the cache."""
        raise NotImplementedError

    def clear(self):
        """Empty the whole cache."""
        raise NotImplementedError


class NullCache(_BaseCache):

    """A cache which does nothing.

    Useful for debugging.
    """

    def put(self, item, *args, **kwargs):
        """Put an item into the cache."""

    def get(self, *args, **kwargs):
        """Get an item from the cache."""
        return None

    def delete(self, *args, **kwargs):
        """Delete an item from the cache."""

    def clear(self):
        """Empty the whole cache."""


class TimedCache(_BaseCache):

    """A simple thread-safe cache for caching method return values.

    The cache key is generated by from the given ``*args`` and ``**kwargs``.
    Items are expired from the cache after a given period of time.

    Example:
        >>> from time import sleep
        >>> cache = TimedCache()
        >>> cache.put("item", 'some', kw='args', timeout=3)
        >>> # Fetch the item again, by providing the same args and kwargs.
        >>> assert cache.get('some', kw='args') == "item"
        >>> # Providing different args or kwargs will not return the item.
        >>> assert not cache.get('some', 'otherargs') == "item"
        >>> # Waiting for less than the provided timeout does not cause the
        >>> # item to expire.
        >>> sleep(2)
        >>> assert cache.get('some', kw='args') == "item"
        >>> # But waiting for longer does.
        >>> sleep(2)
        >>> assert not cache.get('some', kw='args') == "item"

    Warning:
        At present, the cache can theoretically grow and grow, since entries
        are not automatically purged, though in practice this is unlikely
        since there are not that many different combinations of arguments in
        the places where it is used in SoCo, so not that many different
        cache entries will be created. If this becomes a problem,
        use a thread and timer to purge the cache, or rewrite this to use
        LRU logic!
    """

    def __init__(self, default_timeout=0):
        """
        Args:
            default_timeout (int): The default number of seconds after
            which items will be expired.
        """
        super().__init__()
        #: `int`: The default caching expiry interval in seconds.
        self.default_timeout = default_timeout
        # A thread lock for the cache
        self._cache_lock = threading.Lock()

    def get(self, *args, **kwargs):
        """Get an item from the cache for this combination of args and kwargs.

        Args:
            *args: any arguments.
            **kwargs: any keyword arguments.

        Returns:
            object: The object which has been found in the cache, or `None` if
            no unexpired item is found. This means that there is no point
            storing an item in the cache if it is `None`.

        """
        if not self.enabled:
            return None
        # Look in the cache to see if there is an unexpired item. If there is
        # we can just return the cached result.
        cache_key = self.make_key(args, kwargs)
        # Lock and load
        with self._cache_lock:
            if cache_key in self._cache:
                expirytime, item = self._cache[cache_key]

                if expirytime >= time():
                    return item
                else:
                    # An expired item is present - delete it
                    del self._cache[cache_key]
        # Nothing found
        return None

    def put(self, item, *args, **kwargs):
        """Put an item into the cache, for this combination of args and kwargs.

        Args:
            *args: any arguments.
            **kwargs: any keyword arguments. If ``timeout`` is specified as one
                 of the keyword arguments, the item will remain available
                 for retrieval for ``timeout`` seconds. If ``timeout`` is
                 `None` or not specified, the ``default_timeout`` for this
                 cache will be used. Specify a ``timeout`` of 0 (or ensure that
                 the ``default_timeout`` for this cache is 0) if this item is
                 not to be cached.
        """
        if not self.enabled:
            return
        # Check for a timeout keyword, store and remove it.
        timeout = kwargs.pop("timeout", None)
        if timeout is None:
            timeout = self.default_timeout
        cache_key = self.make_key(args, kwargs)
        # Store the item, along with the time at which it will expire
        with self._cache_lock:
            self._cache[cache_key] = (time() + timeout, item)

    def delete(self, *args, **kwargs):
        """Delete an item from the cache for this combination of args and
        kwargs."""
        cache_key = self.make_key(args, kwargs)
        with self._cache_lock:
            try:
                del self._cache[cache_key]
            except KeyError:
                pass

    def clear(self):
        """Empty the whole cache."""
        with self._cache_lock:
            self._cache.clear()

    @staticmethod
    def make_key(*args, **kwargs):
        """Generate a unique, hashable, representation of the args and kwargs.

        Args:
            *args: any arguments.
            **kwargs: any keyword arguments.

        Returns:
            str: the key.
        """
        # This is not entirely straightforward, since args and kwargs may
        # contain mutable items and unicode. Possibilities include using
        # __repr__, frozensets, and code from Py3's LRU cache. But pickle
        # works, and although it is not as fast as some methods, it is good
        # enough at the moment
        cache_key = dumps((args, kwargs))
        return cache_key


class Cache(NullCache):

    """A factory class which returns an instance of a cache subclass.

    A `TimedCache` is returned, unless `config.CACHE_ENABLED` is `False`,
    in which case a `NullCache` will be returned.
    """

    def __new__(cls, *args, **kwargs):
        if config.CACHE_ENABLED:
            new_cls = TimedCache
        else:
            new_cls = NullCache
        instance = super().__new__(new_cls)
        instance.__init__(*args, **kwargs)
        return instance



"""This module is for parsing and conversion functions that needs
objects from both music library and music service data structures

"""


import logging
from urllib.parse import urlparse

from .xml import XML, ns_tag
from .data_structures import didl_class_to_soco_class
from .exceptions import DIDLMetadataError
from .music_services.data_structures import get_class
from .music_services.music_service import desc_from_uri


_LOG = logging.getLogger(__name__)
_LOG.addHandler(logging.NullHandler())
_LOG.debug("%s imported", __name__)


def from_didl_string(string):
    """Convert a unicode xml string to a list of `DIDLObjects <DidlObject>`.

    Args:
        string (str): A unicode string containing an XML representation of one
            or more DIDL-Lite items (in the form  ``'<DIDL-Lite ...>
            ...</DIDL-Lite>'``)

    Returns:
        list: A list of one or more instances of `DidlObject` or a subclass
    """
    items = []
    root = XML.fromstring(string.encode("utf-8"))
    for elt in root:
        if elt.tag.endswith("item") or elt.tag.endswith("container"):
            item_class = elt.findtext(ns_tag("upnp", "class"))
            cls = didl_class_to_soco_class(item_class)
            item = cls.from_element(elt)
            item = attempt_datastructure_upgrade(item)
            items.append(item)
        else:
            # <desc> elements are allowed as an immediate child of <DIDL-Lite>
            # according to the spec, but I have not seen one there in Sonos, so
            # we treat them as illegal. May need to fix this if this
            # causes problems.
            raise DIDLMetadataError("Illegal child of DIDL element: <%s>" % elt.tag)
    _LOG.debug(
        'Created data structures: %.20s (CUT) from Didl string "%.20s" (CUT)',
        items,
        string,
    )
    return items


# Obviously imcomplete, but missing entries will not result in error, but just
# a logged warning and no upgrade of the data structure
DIDL_NAME_TO_QUALIFIED_MS_NAME = {"DidlMusicTrack": "MediaMetadataTrack"}


def attempt_datastructure_upgrade(didl_item):
    """Attempt to upgrade a didl_item to a music services data structure
    if it originates from a music services

    """
    try:
        resource = didl_item.resources[0]
    except IndexError:
        _LOG.debug("Upgrade not possible, no resources")
        return didl_item

    if resource.uri and resource.uri.startswith("x-sonos-http"):
        # Get data
        uri = resource.uri
        # Now we need to create a DIDL item id. It seems to be based on the uri
        path = urlparse(uri).path
        # Strip any extensions, eg .mp3, from the end of the path
        path = path.rsplit(".", 1)[0]
        # The ID has an 8 (hex) digit prefix. But it doesn't seem to
        # matter what it is!
        item_id = "11111111{}".format(path)

        # Pass over all the available metadata in the metadata dict, in the
        # future ask ms data structure to upgrade metadata from the service
        metadata = {}
        for key, value in didl_item.to_dict().items():
            if key not in metadata:
                metadata[key] = value

        # Get class
        try:
            cls = get_class(
                DIDL_NAME_TO_QUALIFIED_MS_NAME[didl_item.__class__.__name__]
            )
        except KeyError:
            # The data structure should be upgraded, but there is an entry
            # missing from DIDL_NAME_TO_QUALIFIED_MS_NAME. Log this as a
            # warning.
            class_name = didl_item.__class__.__name__
            if class_name not in ["DidlFavorite"]:
                _LOG.warning(
                    "DATA STRUCTURE UPGRADE FAIL. Unable to upgrade music library "
                    "data structure to music service data structure because an "
                    "entry is missing for %s in DIDL_NAME_TO_QUALIFIED_MS_NAME. "
                    "This should be reported as a bug.",
                    class_name,
                )
            return didl_item

        upgraded_item = cls(
            item_id=item_id,
            desc=desc_from_uri(resource.uri),
            resources=didl_item.resources,
            uri=uri,
            metadata_dict=metadata,
        )
        _LOG.debug("Item %s upgraded to %s", didl_item, upgraded_item)
        return upgraded_item

    _LOG.debug("Upgrade not necessary")
    return didl_item



"""Tests for the soap module."""


from pysonos.soap import SoapMessage
from pysonos.xml import XML
from unittest import mock


DUMMY_VALID_RESPONSE = "".join(
    [
        '<?xml version="1.0"?>',
        '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        "<s:Body>",
        "<u:GetLEDStateResponse ",
        'xmlns:u="urn:schemas-upnp-org:service:DeviceProperties:1">',
        "<CurrentLEDState>On</CurrentLEDState>",
        "<Unicode>data</Unicode>",
        "</u:GetLEDStateResponse>",
        "</s:Body>",
        "</s:Envelope>",
    ]
)  # noqa PEP8


def test_init():
    """Tests for initialisation of classes."""
    s = SoapMessage("http://endpoint_url", "a_method")
    assert s.endpoint == "http://endpoint_url"
    assert s.method == "a_method"
    assert s.parameters == []
    assert s.soap_action is None
    assert s.http_headers is None
    assert s.soap_header is None
    assert s.namespace is None
    assert s.request_args == {}


def test_prepare_headers():
    """Check http_headers are correctly prepared."""
    s = SoapMessage("endpoint", "method")
    h = s.prepare_headers({"test1": "one", "test2": "two"}, None)
    assert h == {
        "Content-Type": 'text/xml; charset="utf-8"',
        "test2": "two",
        "test1": "one",
    }
    h = s.prepare_headers(
        http_headers={"test1": "one", "test2": "two"}, soap_action="soapaction"
    )
    assert h == {
        "Content-Type": 'text/xml; charset="utf-8"',
        "test2": "two",
        "test1": "one",
        "SOAPACTION": '"soapaction"',
    }


def test_prepare_soap_header():
    """Check that the SOAP header is correctly wrapped."""
    s = SoapMessage("endpoint", "method")
    h = s.prepare_soap_header("<a><b></b></a>")
    assert h == "<s:Header><a><b></b></a></s:Header>"
    s = SoapMessage("endpoint", "method", http_headers={"test1": "one", "test2": "two"})
    h = s.prepare_soap_header(None)
    assert h == ""


def test_prepare_soap_body():
    """Check that the SOAP body is correctly prepared."""
    # No params
    s = SoapMessage("endpoint", "method")
    b = s.prepare_soap_body("a_method", [], None)
    assert b == "<a_method></a_method>"
    # One param
    b = s.prepare_soap_body("a_method", [("one", "1")], None)
    assert b == "<a_method><one>1</one></a_method>"
    # Two params
    b = s.prepare_soap_body("a_method", [("one", "1"), ("two", "2")], None)
    assert b == "<a_method><one>1</one><two>2</two></a_method>"

    # And with a namespace

    b = s.prepare_soap_body(
        "a_method", [("one", "1"), ("two", "2")], "http://a_namespace"
    )
    assert (
        b == "<a_method "
        'xmlns="http://a_namespace"><one>1</one><two>2</two'
        "></a_method>"
    )


def test_prepare():
    """Test preparation of whole SOAP message."""
    s = SoapMessage(
        endpoint="endpoint",
        method="getData",
        parameters=[("one", "1")],
        http_headers={"timeout": "3"},
        soap_action="ACTION",
        soap_header="<a_header>data</a_header>",
        namespace="http://namespace.com",
    )
    headers, data = s.prepare()
    assert (
        data == '<?xml version="1.0"?><s:Envelope '
        'xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" '
        's:encodingStyle="http://schemas.xmlsoap.org/soap'
        '/encoding/"><s:Header><a_header>data</a_header></s'
        ":Header><s:Body><getData "
        'xmlns="http://namespace.com"><one>1</one></getData></s'
        ":Body></s:Envelope>"
    )


def test_call():
    """Calling a command should result in an http request."""

    s = SoapMessage(
        endpoint="http://endpoint.example.com",
        method="getData",
        parameters=[("one", "1")],
        http_headers={"user-agent": "sonos"},
        soap_action="ACTION",
        soap_header="<a_header>data</a_header>",
        namespace="http://namespace.com",
        other_arg=4,
    )

    response = mock.MagicMock()
    response.headers = {}
    response.status_code = 200
    response.content = DUMMY_VALID_RESPONSE
    with mock.patch("requests.post", return_value=response) as fake_post:
        result = s.call()
        assert XML.tostring(result)
        fake_post.assert_called_once_with(
            "http://endpoint.example.com",
            headers={
                "SOAPACTION": '"ACTION"',
                "Content-Type": 'text/xml; charset="utf-8"',
                "user-agent": "sonos",
            },
            data=mock.ANY,
            other_arg=4,
        )



"""Module to test the data structure classes with pytest."""


import pytest

from pysonos import data_structures
from pysonos.exceptions import DIDLMetadataError
from pysonos.xml import XML


def assert_xml_equal(left, right, explain=None):
    """Helper function for comparing XML elements.

    Causes useful information to be output under pytest as to the differences
    between elements

    Args
         left (Element): an Elementtree.Element to compare
         right (Element): an Element to compare it with

    Raises
        AssertionError: if the Elements do not match
    """

    def _build_explanation(left, right, explain):
        if left.tag != right.tag:
            explain.append(
                "tag <{}> does not match tag <{}>".format(left.tag, right.tag)
            )
        for name, value in left.attrib.items():
            if right.get(name) != value:
                explain.append(
                    "%s attribute of element <%s> does not match: %s=%r, %s=%r"
                    % (name, left.tag, name, value, name, right.get(name))
                )
        for name in right.attrib:
            if name not in left.attrib:
                explain.append(
                    "right element <%s> has attribute %s but left does not"
                    % (left.tag, name)
                )
        if left.text != right.text:
            explain.append(
                "text for element <{}>: {!r} != {!r}".format(
                    left.tag, left.text, right.text
                )
            )
        if left.tail != right.tail:
            explain.append(
                "tail for element <{}>: {!r} != {!r}".format(
                    left.tag, left.text, right.text
                )
            )
        for i1, i2 in zip(left, right):
            _build_explanation(i1, i2, explain)
        return

    explain = []
    _build_explanation(left, right, explain)
    if explain != []:
        header = "Comparing XML elements {} and {}".format(left, right)
        assert False, header + "\n".join(explain)


class TestResource:
    """Testing the Resource class."""

    def test_create_didl_resource_with_no_params(self):
        with pytest.raises(TypeError):
            res = data_structures.DidlResource()

    def test_create_didl_resource(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        assert res.uri == "a%20uri"
        assert res.protocol_info == "a:protocol:info:xx"

    def test_create_didl_resource_to_from_element(self, helpers):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx", bitrate=3)
        elt = res.to_element()
        assert helpers.compare_xml(
            elt,
            XML.fromstring(
                b'<res bitrate="3" ' b'protocolInfo="a:protocol:info:xx">a%20uri</res>'
            ),
        )
        assert data_structures.DidlResource.from_element(elt) == res

    def test_didl_resource_to_dict(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        rez = res.to_dict()
        assert rez["uri"] == "a%20uri"
        assert rez["protocol_info"] == "a:protocol:info:xx"
        assert len(rez) == 12

    def test_didl_resource_to_dict_remove_nones(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        rez = res.to_dict(remove_nones=True)
        assert rez["uri"] == "a%20uri"
        assert rez["protocol_info"] == "a:protocol:info:xx"
        assert len(rez) == 2

    def test_didl_resource_from_dict(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        rez = data_structures.DidlResource.from_dict(res.to_dict())
        assert res == rez

    def test_didl_resource_from_dict_remove_nones(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        rez = data_structures.DidlResource.from_dict(res.to_dict(remove_nones=True))
        assert res == rez

    def test_didl_resource_eq(self):
        res = data_structures.DidlResource("a%20uri", "a:protocol:info:xx")
        assert res != data_structures.DidlObject(
            title="a_title", parent_id="pid", item_id="iid"
        )
        assert res is not None
        assert res == res


class TestDidlObject:
    """Testing the DidlObject base class."""

    didl_xml = """
    <item xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
      xmlns:dc="http://purl.org/dc/elements/1.1/"
      xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
      id="iid" parentID="pid" restricted="true">
        <dc:title>the_title</dc:title>
        <upnp:class>object</upnp:class>
        <dc:creator>a_creator</dc:creator>
        <desc id="cdudn"
          nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">DUMMY</desc>
    </item>
    """

    def test_create_didl_object_with_no_params(self):
        with pytest.raises(TypeError):
            didl_object = data_structures.DidlObject()

    def test_create_didl_object_with_disallowed_params(self):
        with pytest.raises(ValueError) as excinfo:
            didl_object = data_structures.DidlObject(
                title="a_title", parent_id="pid", item_id="iid", bad_args="other"
            )
        assert "not allowed" in str(excinfo.value)

    def test_create_didl_object_with_good_params(self):
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            desc="dummy",
        )
        assert didl_object is not None
        assert didl_object.title == "a_title"
        assert didl_object.parent_id == "pid"
        assert didl_object.item_id == "iid"
        assert didl_object.creator == "a_creator"
        assert didl_object.resources == []
        assert didl_object.desc == "dummy"

    def test_didl_object_from_wrong_element(self):
        # Using the wrong element
        elt = XML.fromstring("""<res>URI</res>""")
        with pytest.raises(DIDLMetadataError) as excinfo:
            didl_object = data_structures.DidlObject.from_element(elt)
        assert "Wrong element. Expected <item> or <container>, "
        "got <res> for class object" in str(excinfo.value)

    def test_didl_object_from_element(self):
        elt = XML.fromstring(self.didl_xml)
        didl_object = data_structures.DidlObject.from_element(elt)
        assert didl_object.title == "the_title"
        assert didl_object.parent_id == "pid"
        assert didl_object.item_id == "iid"
        assert didl_object.creator == "a_creator"
        assert didl_object.desc == "DUMMY"
        assert didl_object.item_class == "object"

    def test_didl_object_from_element_unoff_subelement(self):
        """Test that for a DidlObject created from an element with an
        unofficial .# specified sub class, that the sub class is
        completely ignored

        """
        elt = XML.fromstring(self.didl_xml.replace("object", "object.#SubClass"))
        didl_object = data_structures.DidlObject.from_element(elt)
        assert didl_object.item_class == "object"

    def test_didl_object_from_wrong_class(self):
        # mismatched upnp class
        bad_elt1 = XML.fromstring(
            """<item xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
              xmlns:dc="http://purl.org/dc/elements/1.1/"
              xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
              id="iid" parentID="pid" restricted="true">
                 <dc:title>the_title</dc:title>
                 <upnp:class>object.item</upnp:class>
                 <dc:creator>a_creator</dc:creator>
                 <desc id="cdudn"
                   nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
                   RINCON_AssociatedZPUDN
                 </desc>
               </item>
        """
        )
        with pytest.raises(DIDLMetadataError) as excinfo:
            didl_object = data_structures.DidlObject.from_element(bad_elt1)
        assert ("UPnP class is incorrect. Expected 'object', got 'object.item'") in str(
            excinfo.value
        )

    def test_didl_object_from_dict(self):
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            desc="dummy",
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "dummy",
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object
        # adding in an attibute not in _translation should make no difference
        the_dict["creator"] = "another_creator"
        assert data_structures.DidlObject.from_dict(the_dict) != didl_object
        # round trip
        assert data_structures.DidlObject.from_dict(the_dict).to_dict() == the_dict

    def test_didl_object_from_dict_resources(self):
        resources_list = [data_structures.DidlResource("a%20uri", "a:protocol:info:xx")]
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            desc="dummy",
            resources=resources_list,
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "dummy",
            "resources": [resource.to_dict() for resource in resources_list],
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object

    def test_didl_object_from_dict_resources_remove_nones(self):
        resources_list = [data_structures.DidlResource("a%20uri", "a:protocol:info:xx")]
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            desc="dummy",
            resources=resources_list,
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "dummy",
            "resources": [
                resource.to_dict(remove_nones=True) for resource in resources_list
            ],
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object

    def test_didl_comparisons(self):
        didl_object_1 = data_structures.DidlObject(
            title="a_title", parent_id="pid", item_id="iid", creator="a_creator"
        )
        didl_object_2 = data_structures.DidlObject(
            title="a_title", parent_id="pid", item_id="iid", creator="a_creator"
        )
        # should be not the same, but equal!
        assert didl_object_1 is not didl_object_2
        assert didl_object_1 == didl_object_2
        didl_object_3 = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a__different_creator",
        )
        assert didl_object_3 != didl_object_1

    def test_didl_object_to_dict(self):
        didl_object = data_structures.DidlObject(
            title="a_title", parent_id="pid", item_id="iid", creator="a_creator"
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "RINCON_AssociatedZPUDN",
        }
        assert didl_object.to_dict() == the_dict
        # adding in an attibute not in _translation should make no difference
        didl_object.other = "other"
        assert didl_object.to_dict() == the_dict
        # but changing on the other should
        didl_object.creator = "another"
        assert didl_object.to_dict() != the_dict

    def test_didl_object_to_dict_resources(self):
        resources_list = [data_structures.DidlResource("a%20uri", "a:protocol:info:xx")]
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            resources=resources_list,
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "RINCON_AssociatedZPUDN",
            "resources": [resource.to_dict() for resource in resources_list],
        }
        assert didl_object.to_dict() == the_dict

    def test_didl_object_to_dict_resources_remove_nones(self):
        resources_list = [data_structures.DidlResource("a%20uri", "a:protocol:info:xx")]
        didl_object = data_structures.DidlObject(
            title="a_title",
            parent_id="pid",
            item_id="iid",
            creator="a_creator",
            resources=resources_list,
        )
        the_dict = {
            "title": "a_title",
            "parent_id": "pid",
            "item_id": "iid",
            "creator": "a_creator",
            "restricted": True,
            "desc": "RINCON_AssociatedZPUDN",
            "resources": [
                resource.to_dict(remove_nones=True) for resource in resources_list
            ],
        }
        assert didl_object.to_dict(remove_nones=True) == the_dict

    def test_didl_object_to_element(self):
        didl_object = data_structures.DidlObject(
            title="a_title", parent_id="pid", item_id="iid", creator="a_creator"
        )
        # we seem to have to go through this to get ElementTree to deal
        # with namespaces properly!
        elt = XML.fromstring(XML.tostring(didl_object.to_element(True)))
        elt2 = XML.fromstring(
            '<dummy xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
            + 'xmlns:dc="http://purl.org/dc/elements/1.1/" '
            + 'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">'
            + '<item id="iid" parentID="pid" restricted="true">'
            + "<dc:title>a_title</dc:title>"
            + "<dc:creator>a_creator</dc:creator>"
            + '<upnp:class>object</upnp:class><desc id="cdudn" '
            + 'nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">'
            + "RINCON_AssociatedZPUDN</desc></item></dummy>"
        )[0]
        assert_xml_equal(elt2, elt)


# There is an overview of all observed classes, whether they are in or
# out of base DIDL in official_and_extended_didl_classes.txt


def test_didl_object_inheritance():
    """Test that DIDL object inheritance is as indicated by the didl class"""
    class_dict = data_structures._DIDL_CLASS_TO_CLASS.copy()
    class_dict["object"] = data_structures.DidlObject
    for didl_class, soco_class in data_structures._DIDL_CLASS_TO_CLASS.items():
        # Skip this one, because its DIDL class is expected to be an error
        if didl_class == "object.itemobject.item.sonos-favorite":
            continue
        # object does not inherit
        if didl_class == "object":
            continue

        # First make sure it is registered with the correct DIDL class
        assert didl_class == soco_class.item_class

        base_didl_class = ".".join(didl_class.split(".")[:-1])
        base_class = data_structures._DIDL_CLASS_TO_CLASS[base_didl_class]
        assert base_class == soco_class.__bases__[0]



# pylint: disable=star-args,too-many-locals

"""Plugin for the Wimp music service (Service ID 20)"""


import locale
import socket

import requests

from ..exceptions import SoCoUPnPException, UnknownXMLStructure
from ..ms_data_structures import (
    MSAlbum,
    MSAlbumList,
    MSArtist,
    MSArtistTracklist,
    MSCollection,
    MSFavorites,
    MSPlaylist,
    MSTrack,
    get_ms_item,
)
from ..services import MusicServices
from ..utils import really_utf8
from ..xml import XML
from .__init__ import SoCoPlugin

__all__ = ["Wimp"]


def _post(url, headers, body, retries=3, timeout=3.0):
    """Try 3 times to request the content.

    :param headers: The HTTP headers
    :type headers: dict
    :param body: The body of the HTTP post
    :type body: str
    :param retries: The number of times to retry before giving up
    :type retries: int
    :param timeout: The time to wait for the post to complete, before timing
        out
    :type timeout: float
    """
    retry = 0
    out = None
    while out is None:
        try:
            out = requests.post(url, headers=headers, data=body, timeout=timeout)
        # Due to a bug in requests, the post command will sometimes fail to
        # properly wrap a socket.timeout exception in requests own exception.
        # See https://github.com/kennethreitz/requests/issues/2045
        # Until this is fixed, we need to catch both types of exceptions
        except (requests.exceptions.Timeout, socket.timeout) as exception:
            retry += 1
            if retry == retries:
                # pylint: disable=maybe-no-member
                raise requests.exceptions.Timeout(exception.message)
    return out


def _ns_tag(ns_id, tag):
    """Return a namespace/tag item. The ns_id is translated to a full name
    space via the NS module variable.

    :param ns_id: The name space ID. Translated to a namespace via the module
        variable NS
    :type ns_id: str
    :param tag: The tag
    :type str: str
    """
    return "{{{}}}{}".format(NS[ns_id], tag)


def _get_header(soap_action):
    """Return the HTTP for SOAP Action.

    :param soap_action: The soap action to include in the header. Can be either
        'search' or 'get_metadata'
    :type soap_action: str
    """
    # This way of setting accepted language is obviously flawed, in that it
    # depends on the locale settings of the system. However, I'm unsure if
    # they are actually used. The character coding is set elsewhere and I think
    # the available music in each country is bound to the account.
    language, _ = locale.getdefaultlocale()
    if language is None:
        language = ""
    else:
        language = language.replace("_", "-") + ", "

    header = {
        "CONNECTION": "close",
        "ACCEPT-ENCODING": "gzip",
        "ACCEPT-LANGUAGE": "{}en-US;q=0.9".format(language),
        "Content-Type": 'text/xml; charset="utf-8"',
        "SOAPACTION": SOAP_ACTION[soap_action],
    }
    return header


class Wimp(SoCoPlugin):

    """Class that implements a Wimp plugin.

    Note:
        There is an (apparent) in-consistency in the use of one data
        type from the Wimp service. When searching for playlists, the XML
        returned by the Wimp server indicates, that the type is an 'album
        list', and it thus suggest, that this type is used for a list of
        tracks (as expected for a playlist), and this data type is reported
        to be playable. However, when browsing the music tree, the Wimp
        server will return items of 'album list' type, but in this case it
        is used for a list of albums and it is not playable. This plugin
        maintains this (apparent) in-consistency to stick as close to the
        reported data as possible, so search for playlists returns
        MSAlbumList that are playable and while browsing the content tree
        the MSAlbumList items returned to you are not playable.


    Note:
       Wimp in some cases lists tracks that are not available. In these
       cases, while it will correctly report these tracks as not being
       playable, the containing data structure like e.g. the album they are
       on may report that they are playable. Trying to add one of these to
       the queue will return a SoCoUPnPException with error code '802'.

    """

    def __init__(self, soco, username, retries=3, timeout=3.0):
        """Initialize the plugin.

        :param soco: The soco instance to retrieve the session ID for the music
            service
        :type: :py:class:`soco.SoCo`
        :param username: The username for the music service
        :type username: str
        :param retries: The number of times to retry before giving up
        :type retries: int
        :param timeout: The time to wait for the post to complete, before
            timing out. The Wimp server seems either slow to respond or to
            make the queries internally, so the timeout should probably not be
            shorter than 3 seconds.
        :type timeout: float

        Note:

            If you are using a phone number as the username and are
            experiencing problems connecting, then try to prepend the area
            code (no + or 00). I.e. if your phone number is 12345678 and you
            are from denmark, then use 4512345678. This must be set up the
            same way in the Sonos device.  For details see `here
            <https://wimp.zendesk.com/hc/da/articles/204311810-Hvorfor-kan
            -jeg-ikke-logge-p%C3%A5-WiMP-med-min-Sonos-n%C3%A5r-jeg-har-et
            -gyldigt-abonnement->`_ (In Danish)
        """
        super().__init__(soco)

        # Instantiate variables
        self._url = "http://client.wimpmusic.com/sonos/services/Sonos"
        self._serial_number = soco.get_speaker_info()["serial_number"]
        self._username = username
        self._service_id = 20
        self._http_vars = {"retries": retries, "timeout": timeout}

        # Get a session id for the searches
        self._music_services = MusicServices(soco)
        response = self._music_services.GetSessionId(
            [("ServiceId", 20), ("Username", username)]
        )
        self._session_id = response["SessionId"]

    @property
    def name(self):
        """Return the human read-able name for the plugin"""
        return "Wimp Plugin for {}".format(self._username)

    @property
    def username(self):
        """Return the username."""
        return self._username

    @property
    def service_id(self):
        """Return the service id."""
        return self._service_id

    @property
    def description(self):
        """Return the music service description for the DIDL metadata on the
        form ``'SA_RINCON5127_...self.username...'``"""
        return "SA_RINCON5127_{}".format(self._username)

    def get_tracks(self, search, start=0, max_items=100):
        """Search for tracks.

        See get_music_service_information for details on the arguments
        """
        return self.get_music_service_information("tracks", search, start, max_items)

    def get_albums(self, search, start=0, max_items=100):
        """Search for albums.

        See get_music_service_information for details on the arguments
        """
        return self.get_music_service_information("albums", search, start, max_items)

    def get_artists(self, search, start=0, max_items=100):
        """Search for artists.

        See get_music_service_information for details on the arguments
        """
        return self.get_music_service_information("artists", search, start, max_items)

    def get_playlists(self, search, start=0, max_items=100):
        """Search for playlists.

        See get_music_service_information for details on the arguments.

        Note:

            Un-intuitively this method returns MSAlbumList items. See
            note in class doc string for details.
        """
        return self.get_music_service_information("playlists", search, start, max_items)

    def get_music_service_information(
        self, search_type, search, start=0, max_items=100
    ):
        """Search for music service information items.

        :param search_type: The type of search to perform, possible values are:
            'artists', 'albums', 'tracks' and 'playlists'
        :type search_type: str
        :param search: The search string to use
        :type search: str
        :param start: The starting index of the returned items
        :type start: int
        :param max_items: The maximum number of returned items
        :type max_items: int

        Note:
            Un-intuitively the playlist search returns MSAlbumList
            items. See note in class doc string for details.
        """
        # Check input
        if search_type not in ["artists", "albums", "tracks", "playlists"]:
            message = "The requested search {} is not valid".format(search_type)
            raise ValueError(message)
        # Transform search: tracks -> tracksearch
        search_type = "{}earch".format(search_type)
        parent_id = SEARCH_PREFIX.format(search_type=search_type, search=search)

        # Perform search
        body = self._search_body(search_type, search, start, max_items)
        headers = _get_header("search")
        response = _post(self._url, headers, body, **self._http_vars)
        self._check_for_errors(response)
        result_dom = XML.fromstring(response.text.encode("utf-8"))

        # Parse results
        search_result = result_dom.find(".//" + _ns_tag("", "searchResult"))
        out = {"item_list": []}
        for element in ["index", "count", "total"]:
            out[element] = search_result.findtext(_ns_tag("", element))

        if search_type == "tracksearch":
            item_name = "mediaMetadata"
        else:
            item_name = "mediaCollection"
        for element in search_result.findall(_ns_tag("", item_name)):
            out["item_list"].append(get_ms_item(element, self, parent_id))

        return out

    def browse(self, ms_item=None):
        """Return the sub-elements of item or of the root if item is None

        :param item: Instance of sub-class of
            :py:class:`soco.data_structures.MusicServiceItem`. This object must
            have item_id, service_id and extended_id properties

        Note:
            Browsing a MSTrack item will return itself.

        Note:
            This plugin cannot yet set the parent ID of the results
            correctly when browsing
            :py:class:`soco.data_structures.MSFavorites` and
            :py:class:`soco.data_structures.MSCollection` elements.

        """
        # Check for correct service
        if ms_item is not None and ms_item.service_id != self._service_id:
            message = "This music service item is not for this service"
            raise ValueError(message)

        # Form HTTP body and set parent_id
        if ms_item:
            body = self._browse_body(ms_item.item_id)
            parent_id = ms_item.extended_id
            if parent_id is None:
                parent_id = ""
        else:
            body = self._browse_body("root")
            parent_id = "0"

        # Get HTTP header and post
        headers = _get_header("get_metadata")
        response = _post(self._url, headers, body, **self._http_vars)

        # Check for errors and get XML
        self._check_for_errors(response)
        result_dom = XML.fromstring(really_utf8(response.text))
        # Find the getMetadataResult item ...
        xpath_search = ".//" + _ns_tag("", "getMetadataResult")
        metadata_result = list(result_dom.findall(xpath_search))
        # ... and make sure there is exactly 1
        if len(metadata_result) != 1:
            raise UnknownXMLStructure(
                "The results XML has more than 1 'getMetadataResult'. This "
                "is unexpected and parsing will dis-continue."
            )
        metadata_result = metadata_result[0]

        # Browse the children of metadata result
        out = {"item_list": []}
        for element in ["index", "count", "total"]:
            out[element] = metadata_result.findtext(_ns_tag("", element))
        for result in metadata_result:
            if result.tag in [
                _ns_tag("", "mediaCollection"),
                _ns_tag("", "mediaMetadata"),
            ]:
                out["item_list"].append(get_ms_item(result, self, parent_id))
        return out

    @staticmethod
    def id_to_extended_id(item_id, item_class):
        """Return the extended ID from an ID.

        :param item_id: The ID of the music library item
        :type item_id: str
        :param cls: The class of the music service item
        :type cls: Sub-class of
            :py:class:`soco.data_structures.MusicServiceItem`

        The extended id can be something like 00030020trackid_22757082
        where the id is just trackid_22757082. For classes where the prefix is
        not known returns None.
        """
        out = ID_PREFIX[item_class]
        if out:
            out += item_id
        return out

    @staticmethod
    def form_uri(item_content, item_class):
        """Form the URI for a music service element.

        :param item_content: The content dict of the item
        :type item_content: dict
        :param item_class: The class of the item
        :type item_class: Sub-class of
            :py:class:`soco.data_structures.MusicServiceItem`
        """
        extension = None
        if "mime_type" in item_content:
            extension = MIME_TYPE_TO_EXTENSION[item_content["mime_type"]]
        out = URIS.get(item_class)
        if out:
            out = out.format(extension=extension, **item_content)
        return out

    def _search_body(self, search_type, search_term, start, max_items):
        """Return the search XML body.

        :param search_type: The search type
        :type search_type: str
        :param search_term: The search term e.g. 'Jon Bon Jovi'
        :type search_term: str
        :param start: The start index of the returned results
        :type start: int
        :param max_items: The maximum number of returned results
        :type max_items: int

        The XML is formed by adding, to the envelope of the XML returned by
        ``self._base_body``, the following ``Body`` part:

        .. code :: xml

         <s:Body>
           <search xmlns="http://www.sonos.com/Services/1.1">
             <id>search_type</id>
             <term>search_term</term>
             <index>start</index>
             <count>max_items</count>
           </search>
         </s:Body>
        """
        xml = self._base_body()

        # Add the Body part
        XML.SubElement(xml, "s:Body")
        item_attrib = {"xmlns": "http://www.sonos.com/Services/1.1"}
        search = XML.SubElement(xml[1], "search", item_attrib)
        XML.SubElement(search, "id").text = search_type
        XML.SubElement(search, "term").text = search_term
        XML.SubElement(search, "index").text = str(start)
        XML.SubElement(search, "count").text = str(max_items)

        return XML.tostring(xml)

    def _browse_body(self, search_id):
        """Return the browse XML body.

        The XML is formed by adding, to the envelope of the XML returned by
        ``self._base_body``, the following ``Body`` part:

        .. code :: xml

         <s:Body>
           <getMetadata xmlns="http://www.sonos.com/Services/1.1">
             <id>root</id>
             <index>0</index>
             <count>100</count>
           </getMetadata>
         </s:Body>

        .. note:: The XML contains index and count, but the service does not
        seem to respect them, so therefore they have not been included as
        arguments.
        """
        xml = self._base_body()

        # Add the Body part
        XML.SubElement(xml, "s:Body")
        item_attrib = {"xmlns": "http://www.sonos.com/Services/1.1"}
        search = XML.SubElement(xml[1], "getMetadata", item_attrib)
        XML.SubElement(search, "id").text = search_id
        # Investigate this index, count stuff more
        XML.SubElement(search, "index").text = "0"
        XML.SubElement(search, "count").text = "100"

        return XML.tostring(xml)

    def _base_body(self):
        """Return the base XML body, which has the following form:

        .. code :: xml

         <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">
           <s:Header>
             <credentials xmlns="http://www.sonos.com/Services/1.1">
               <sessionId>self._session_id</sessionId>
               <deviceId>self._serial_number</deviceId>
               <deviceProvider>Sonos</deviceProvider>
             </credentials>
           </s:Header>
         </s:Envelope>
        """
        item_attrib = {
            "xmlns:s": "http://schemas.xmlsoap.org/soap/envelope/",
        }
        xml = XML.Element("s:Envelope", item_attrib)

        # Add the Header part
        XML.SubElement(xml, "s:Header")
        item_attrib = {"xmlns": "http://www.sonos.com/Services/1.1"}
        credentials = XML.SubElement(xml[0], "credentials", item_attrib)
        XML.SubElement(credentials, "sessionId").text = self._session_id
        XML.SubElement(credentials, "deviceId").text = self._serial_number
        XML.SubElement(credentials, "deviceProvider").text = "Sonos"

        return xml

    def _check_for_errors(self, response):
        """Check a response for errors.

        :param response: the response from requests.post()
        """
        if response.status_code != 200:
            xml_error = really_utf8(response.text)
            error_dom = XML.fromstring(xml_error)
            fault = error_dom.find(".//" + _ns_tag("s", "Fault"))
            error_description = fault.find("faultstring").text
            error_code = EXCEPTION_STR_TO_CODE[error_description]
            message = "UPnP Error {} received: {} from {}".format(
                error_code, error_description, self._url
            )
            raise SoCoUPnPException(
                message=message,
                error_code=error_code,
                error_description=error_description,
                error_xml=really_utf8(response.text),
            )


SOAP_ACTION = {
    "get_metadata": '"http://www.sonos.com/Services/1.1#getMetadata"',
    "search": '"http://www.sonos.com/Services/1.1#search"',
}
# Note UPnP exception 802 while trying to add a Wimp track indicates that these
# are tracks that not available in Wimp. Do something with that.
EXCEPTION_STR_TO_CODE = {"unknown": 20000, "ItemNotFound": 20001}
SEARCH_PREFIX = "00020064{search_type}:{search}"
ID_PREFIX = {
    MSTrack: "00030020",
    MSAlbum: "0004002c",
    MSArtist: "10050024",
    MSAlbumList: "000d006c",
    MSPlaylist: "0006006c",
    MSArtistTracklist: "100f006c",
    MSFavorites: None,  # This one is unknown
    MSCollection: None,  # This one is unknown
}
MIME_TYPE_TO_EXTENSION = {"audio/aac": "mp4"}
URIS = {
    MSTrack: "x-sonos-http:{item_id}.{extension}?sid={service_id}&flags=32",
    MSAlbum: "x-rincon-cpcontainer:{extended_id}",
    MSAlbumList: "x-rincon-cpcontainer:{extended_id}",
    MSPlaylist: "x-rincon-cpcontainer:{extended_id}",
    MSArtistTracklist: "x-rincon-cpcontainer:{extended_id}",
}
NS = {
    "s": "http://schemas.xmlsoap.org/soap/envelope/",
    "": "http://www.sonos.com/Services/1.1",
}



"""Tests for the services module."""

# These tests require pytest.


import pytest

from pysonos.exceptions import SoCoUPnPException
from pysonos.services import Service, Action, Argument, Vartype

from unittest import mock

# Dummy known-good errors/responses etc.  These are not necessarily valid as
# actual commands, but are valid XML/UPnP. They also contain unicode characters
# to test unicode handling.

DUMMY_ERROR = "".join(
    [
        '<?xml version="1.0"?>',
        "<s:Envelope ",
        'xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" ',
        's:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        "<s:Body>",
        "<s:Fault>",
        "<faultcode>s:Client</faultcode>",
        "<faultstring>UPnPError</faultstring>",
        "<detail>",
        '<UPnPError xmlns="urn:schemas-upnp-org:control-1-0">',
        "<errorCode>607</errorCode>",
        "<errorDescription>Oops μИⅠℂ☺ΔЄ💋</errorDescription>",
        "</UPnPError>",
        "</detail>",
        "</s:Fault>",
        "</s:Body>",
        "</s:Envelope>",
    ]
)  # noqa PEP8

DUMMY_VALID_RESPONSE = "".join(
    [
        '<?xml version="1.0"?>',
        '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        "<s:Body>",
        "<u:GetLEDStateResponse ",
        'xmlns:u="urn:schemas-upnp-org:service:DeviceProperties:1">',
        "<CurrentLEDState>On</CurrentLEDState>",
        "<Unicode>μИⅠℂ☺ΔЄ💋</Unicode>",
        "</u:GetLEDStateResponse>",
        "</s:Body>",
        "</s:Envelope>",
    ]
)  # noqa PEP8

DUMMY_VALID_ACTION = "".join(
    [
        '<?xml version="1.0"?>',
        '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        "<s:Body>",
        "<u:SetAVTransportURI ",
        'xmlns:u="urn:schemas-upnp-org:service:Service:1">',
        "<InstanceID>0</InstanceID>",
        "<CurrentURI>URI</CurrentURI>",
        "<CurrentURIMetaData></CurrentURIMetaData>",
        "<Unicode>μИⅠℂ☺ΔЄ💋</Unicode>" "</u:SetAVTransportURI>",
        "</s:Body>" "</s:Envelope>",
    ]
)  # noqa PEP8

DUMMY_VARTYPE = Vartype("string", None, None, None)

DUMMY_ACTIONS = [
    Action(
        name="Test",
        in_args=[
            Argument(name="Argument1", vartype=DUMMY_VARTYPE),
            Argument(name="Argument2", vartype=DUMMY_VARTYPE),
        ],
        out_args=[],
    )
]

DUMMY_ARGS = [("Argument1", 1), ("Argument2", 2)]

DUMMY_ARGS_ALTERNATIVE = [("Argument1", 3), ("Argument2", 2)]


@pytest.fixture()
def service():
    """A mock Service, for use as a test fixture."""

    mock_soco = mock.MagicMock()
    mock_soco.ip_address = "192.168.1.101"
    mock_service = Service(mock_soco)
    return mock_service


def test_init_defaults(service):
    """Check default properties are set up correctly."""
    assert service.service_type == "Service"
    assert service.version == 1
    assert service.service_id == "Service"
    assert service.base_url == "http://192.168.1.101:1400"
    assert service.control_url == "/Service/Control"
    assert service.scpd_url == "/xml/Service1.xml"
    assert service.event_subscription_url == "/Service/Event"


def test_method_dispatcher_function_creation(service):
    """Testing __getattr__ functionality."""
    import inspect

    # There should be no testing method
    assert "testing" not in service.__dict__.keys()
    # but we should be able to inspect it
    assert inspect.ismethod(service.testing)
    # and then, having examined it, the method should be cached on the instance
    assert "testing" in service.__dict__.keys()
    assert service.testing.__name__ == "testing"
    # check that send_command is actually called when we invoke a method
    service.send_command = lambda x, y: "Hello {}".format(x)
    assert service.testing(service) == "Hello testing"


def test_method_dispatcher_arg_count(service):
    """_dispatcher should pass its args to send_command."""
    service.send_command = mock.Mock()
    # http://bugs.python.org/issue7688
    # __name__ must be a string in python 2
    method = service.__getattr__("test")
    assert method("onearg")
    service.send_command.assert_called_with("test", "onearg")
    assert method()  # no args
    service.send_command.assert_called_with("test")
    assert method("one", cache_timeout=4)  # one arg + cache_timeout
    service.send_command.assert_called_with("test", "one", cache_timeout=4)


def test_wrap(service):
    """wrapping args in XML properly."""
    assert (
        service.wrap_arguments([("first", "one"), ("second", 2)])
        == "<first>one</first><second>2</second>"
    )
    assert service.wrap_arguments() == ""
    # Unicode
    assert (
        service.wrap_arguments([("unicode", "μИⅠℂ☺ΔЄ💋")])
        == "<unicode>μИⅠℂ☺ΔЄ💋</unicode>"
    )
    # XML escaping - do we also need &apos; ?
    assert (
        service.wrap_arguments([("weird", '&<"2')]) == "<weird>&amp;&lt;&quot;2</weird>"
    )


def test_unwrap(service):
    """unwrapping args from XML."""
    assert service.unwrap_arguments(DUMMY_VALID_RESPONSE) == {
        "CurrentLEDState": "On",
        "Unicode": "μИⅠℂ☺ΔЄ💋",
    }


def test_unwrap_invalid_char(service):
    """Test unwrapping args from XML with invalid char"""
    responce_with_invalid_char = DUMMY_VALID_RESPONSE.replace("μИⅠℂ☺ΔЄ💋", "AB")
    # Note, the invalid ^D (code point 0x04) should be filtered out
    assert service.unwrap_arguments(responce_with_invalid_char) == {
        "CurrentLEDState": "On",
        "Unicode": "AB",
    }


def test_compose(service):
    """Test argument composition."""
    service._actions = DUMMY_ACTIONS
    service.DEFAULT_ARGS = {}

    # Detect unknown action
    with pytest.raises(AttributeError):
        service.compose_args("Error", {})
    # Detect missing / unknown arguments
    with pytest.raises(ValueError):
        service.compose_args("Test", {"Argument1": 1})
    with pytest.raises(ValueError):
        service.compose_args("Test", dict(DUMMY_ARGS + [("Error", 3)]))

    # Check correct output
    assert service.compose_args("Test", dict(DUMMY_ARGS)) == DUMMY_ARGS

    # Set Argument1 = 1 as default
    service.DEFAULT_ARGS = dict(DUMMY_ARGS[:1])

    # Check that arguments are completed with default values
    assert service.compose_args("Test", dict(DUMMY_ARGS[1:])) == DUMMY_ARGS
    # Check that given arguments override the default values
    assert (
        service.compose_args("Test", dict(DUMMY_ARGS_ALTERNATIVE))
        == DUMMY_ARGS_ALTERNATIVE
    )


def test_build_command(service):
    """Test creation of SOAP body and headers from a command."""
    headers, body = service.build_command(
        "SetAVTransportURI",
        [
            ("InstanceID", 0),
            ("CurrentURI", "URI"),
            ("CurrentURIMetaData", ""),
            ("Unicode", "μИⅠℂ☺ΔЄ💋"),
        ],
    )
    assert body == DUMMY_VALID_ACTION
    assert headers == {
        "Content-Type": 'text/xml; charset="utf-8"',
        "SOAPACTION": "urn:schemas-upnp-org:service:Service:1#SetAVTransportURI",
    }


def test_send_command(service):
    """Calling a command should result in a http request, unless the cache is
    hit."""
    response = mock.MagicMock()
    response.headers = {}
    response.status_code = 200
    response.text = DUMMY_VALID_RESPONSE
    with mock.patch("requests.post", return_value=response) as fake_post:
        result = service.send_command(
            "SetAVTransportURI",
            [
                ("InstanceID", 0),
                ("CurrentURI", "URI"),
                ("CurrentURIMetaData", ""),
                ("Unicode", "μИⅠℂ☺ΔЄ💋"),
            ],
            cache_timeout=2,
        )
        assert result == {"CurrentLEDState": "On", "Unicode": "μИⅠℂ☺ΔЄ💋"}
        fake_post.assert_called_once_with(
            "http://192.168.1.101:1400/Service/Control",
            headers=mock.ANY,
            data=DUMMY_VALID_ACTION.encode("utf-8"),
            timeout=20,
        )
        # Now the cache should be primed, so try it again
        fake_post.reset_mock()
        result = service.send_command(
            "SetAVTransportURI",
            [
                ("InstanceID", 0),
                ("CurrentURI", "URI"),
                ("CurrentURIMetaData", ""),
                ("Unicode", "μИⅠℂ☺ΔЄ💋"),
            ],
            cache_timeout=0,
        )
        # The cache should be hit, so there should be no http request
        assert not fake_post.called
        # but this should not affefct a call with different params
        fake_post.reset_mock()
        result = service.send_command(
            "SetAVTransportURI",
            [
                ("InstanceID", 1),
                ("CurrentURI", "URI2"),
                ("CurrentURIMetaData", "abcd"),
                ("Unicode", "μИⅠℂ☺ΔЄ💋"),
            ],
        )
        assert fake_post.called
        # calling again after the time interval will avoid the cache
        fake_post.reset_mock()
        import time

        time.sleep(2)
        result = service.send_command(
            "SetAVTransportURI",
            [
                ("InstanceID", 0),
                ("CurrentURI", "URI"),
                ("CurrentURIMetaData", ""),
                ("Unicode", "μИⅠℂ☺ΔЄ💋"),
            ],
        )
        assert fake_post.called


def test_handle_upnp_error(service):
    """Check errors are extracted properly."""
    with pytest.raises(SoCoUPnPException) as E:
        service.handle_upnp_error(DUMMY_ERROR)
    assert (
        "UPnP Error 607 received: Signature Failure from 192.168.1.101"
        == E.value.message
    )
    assert E.value.error_code == "607"
    assert E.value.error_description == "Signature Failure"
    # TODO: Try this with a None Error Code


# TODO: test iter_actions
